import math
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import List

import pytest

from kloppy import hawkeye
from kloppy.domain import (
    DatasetType,
    HawkEyeCoordinateSystem,
    Orientation,
    Point,
    Point3D,
    Provider,
    TrackingDataset,
)


@pytest.fixture(scope="session")
def ball_feeds(base_dir: Path) -> List[Path]:
    return [
        base_dir / "files" / "hawkeye_1_1.football.samples.ball",
        base_dir / "files" / "hawkeye_2_46.football.samples.ball",
    ]


@pytest.fixture(scope="session")
def player_centroid_feeds(base_dir: Path) -> List[Path]:
    return [
        base_dir / "files" / "hawkeye_1_1.football.samples.centroids",
        base_dir / "files" / "hawkeye_2_46.football.samples.centroids",
    ]


@pytest.fixture(scope="session")
def meta_data_xml(base_dir: Path) -> Path:
    return base_dir / "files" / "hawkeye_meta.xml"


@pytest.fixture(scope="session")
def meta_data_json(base_dir: Path) -> Path:
    return base_dir / "files" / "hawkeye_meta.json"


class TestsHawkEyeInputs:
    """Tests related to the various input options."""

    def test_deserialize_single_file(
        self, ball_feeds: List[Path], player_centroid_feeds: List[Path]
    ):
        dataset = hawkeye.load(
            ball_feeds=ball_feeds[0],
            player_centroid_feeds=player_centroid_feeds[0],
            coordinates="hawkeye",
        )
        assert len(dataset) == 3000

    def test_deserialize_multiple_files(
        self,
        ball_feeds: List[Path],
        player_centroid_feeds: List[Path],
    ):
        dataset = hawkeye.load(
            ball_feeds=ball_feeds,
            player_centroid_feeds=player_centroid_feeds,
            coordinates="hawkeye",
        )
        assert len(dataset) == 6000

    def test_deserialize_directory(self, base_dir: Path):
        dataset = hawkeye.load(
            ball_feeds=(base_dir / "files").resolve(),
            player_centroid_feeds=(base_dir / "files").resolve(),
            coordinates="hawkeye",
        )
        assert len(dataset) == 6000

    def test_limit(
        self, ball_feeds: List[Path], player_centroid_feeds: List[Path]
    ):
        dataset = hawkeye.load(
            ball_feeds=ball_feeds,
            player_centroid_feeds=player_centroid_feeds,
            coordinates="hawkeye",
            limit=10,
        )
        assert len(dataset) == 10

    def test_sample_rate(
        self, ball_feeds: List[Path], player_centroid_feeds: List[Path]
    ):
        dataset = hawkeye.load(
            ball_feeds=ball_feeds,
            player_centroid_feeds=player_centroid_feeds,
            coordinates="hawkeye",
            sample_rate=1 / 2,
            limit=300,
        )
        assert len(dataset) == 300
        assert dataset.metadata.frame_rate == 50
        assert dataset.records[0].timestamp.total_seconds() == 0.000080
        assert dataset.records[1].timestamp.total_seconds() == 0.040078

    def test_overwrite_metadata(
        self,
        ball_feeds: List[Path],
        player_centroid_feeds: List[Path],
        meta_data_json: Path,
    ):
        dataset = hawkeye.load(
            ball_feeds=ball_feeds,
            player_centroid_feeds=player_centroid_feeds,
            pitch_length=120.0,
            pitch_width=70.0,
            coordinates="hawkeye",
        )
        assert dataset.metadata.coordinate_system.pitch_length == 120.0
        assert dataset.metadata.coordinate_system.pitch_width == 70.0

        dataset = hawkeye.load(
            ball_feeds=ball_feeds,
            player_centroid_feeds=player_centroid_feeds,
            meta_data=meta_data_json,
            pitch_length=120.0,
            pitch_width=70.0,
            coordinates="hawkeye",
        )
        assert dataset.metadata.coordinate_system.pitch_length == 104.0
        assert dataset.metadata.coordinate_system.pitch_width == 67.0


class TestHawkEyeDeserializer:
    """Tests related to checking the correctness of the deserialization."""

    @pytest.fixture(scope="class")
    def dataset(
        self, ball_feeds: List[Path], player_centroid_feeds: List[Path]
    ) -> TrackingDataset:
        dataset = hawkeye.load(
            ball_feeds=ball_feeds,
            player_centroid_feeds=player_centroid_feeds,
            pitch_length=105.0,
            pitch_width=68.0,
            coordinates="hawkeye",
        )
        assert len(dataset) == 6000
        assert dataset.dataset_type == DatasetType.TRACKING
        return dataset

    def test_provider(self, dataset: TrackingDataset):
        """It should set the provider"""
        assert dataset.metadata.provider == Provider.HAWKEYE

    def test_orientation(self, dataset):
        """It should set the correct orientation"""
        assert dataset.metadata.orientation == Orientation.HOME_AWAY

    def test_frame_rate(self, dataset):
        """It should set the frame rate to 25"""
        assert dataset.metadata.frame_rate == 50

    def test_teams(self, dataset):
        """It should create the teams and player objects"""
        # There should be two teams with the correct names
        assert dataset.metadata.teams[0].name == "Team A"
        assert dataset.metadata.teams[0].starting_formation is None
        assert dataset.metadata.teams[1].name == "Team B"
        assert dataset.metadata.teams[1].starting_formation is None
        # The teams should have the correct players
        player = dataset.metadata.teams[0].get_player_by_id("487487")
        assert player.player_id == "487487"
        assert player.jersey_no == 4
        assert str(player) == "away_4"

    def test_periods(self, dataset):
        """It should create the periods"""
        assert dataset.metadata.periods[0].id == 1
        assert len(dataset.metadata.periods) == 2

        assert dataset.metadata.periods[
            0
        ].start_timestamp == datetime.strptime(
            "2024-09-22T21:00:49.383Z", "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        assert dataset.metadata.periods[0].end_timestamp == datetime.strptime(
            "2024-09-22T21:47:30.592Z", "%Y-%m-%dT%H:%M:%S.%fZ"
        )

        assert dataset.metadata.periods[
            1
        ].start_timestamp == datetime.strptime(
            "2024-09-22T22:03:13.540Z", "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        assert dataset.metadata.periods[1].end_timestamp == datetime.strptime(
            "2024-09-22T23:00:26.296Z", "%Y-%m-%dT%H:%M:%S.%fZ"
        )

    def test_coordinate_system(self, dataset):
        """It should set the correct coordinate system"""
        assert isinstance(
            dataset.metadata.coordinate_system, HawkEyeCoordinateSystem
        )

    def test_pitch_dimensions(self, dataset):
        """It should set the correct pitch dimensions"""
        pitch_dimensions = dataset.metadata.pitch_dimensions
        assert pitch_dimensions.x_dim.min == -52.5
        assert pitch_dimensions.x_dim.max == 52.5
        assert pitch_dimensions.y_dim.min == -34.0
        assert pitch_dimensions.y_dim.max == 34.0

    def test_timestamps(self, dataset):
        """It should set the correct timestamps"""
        assert math.isclose(
            dataset.records[0].timestamp.total_seconds(),
            timedelta(seconds=8e-05).total_seconds(),
            rel_tol=1e-5,
        )
        assert math.isclose(
            dataset.records[400].timestamp.total_seconds(),
            timedelta(seconds=8.019669).total_seconds(),
            rel_tol=1e-5,
        )
        # second frame second half
        assert math.isclose(
            dataset.records[3001].timestamp.total_seconds(),
            timedelta(seconds=0.02008).total_seconds(),
            rel_tol=1e-5,
        )

    def test_ball_coordinates(self, dataset):
        """It should set the correct ball coordinates"""
        assert dataset.records[1].ball_coordinates == Point3D(
            x=0.2773038337512226, y=-0.06614051219344752, z=0.21062164074548867
        )

    def test_player_coordinates(self, dataset):
        """It should set the correct player coordinates"""
        home_player = dataset.metadata.teams[0].players[3]
        assert home_player.player_id == "487487"
        assert dataset.records[0].players_coordinates[home_player] == Point(
            x=21.757139205932617, y=-4.809798240661621
        )

        away_player = dataset.metadata.teams[1].players[3]
        assert away_player.player_id == "487721"
        assert dataset.records[0].players_coordinates[away_player] == Point(
            x=-0.41805464029312134, y=-9.460993766784668
        )


def test_xml_metadata(
    ball_feeds: List[Path],
    player_centroid_feeds: List[Path],
    meta_data_xml: Path,
):
    dataset = hawkeye.load(
        ball_feeds=ball_feeds,
        player_centroid_feeds=player_centroid_feeds,
        meta_data=meta_data_xml,
        coordinates="hawkeye",
    )
    # Check pitch dimensions
    pitch_dimensions = dataset.metadata.pitch_dimensions
    assert pitch_dimensions.x_dim.min == -53.0
    assert pitch_dimensions.x_dim.max == 53.0
    assert pitch_dimensions.y_dim.min == -34.5
    assert pitch_dimensions.y_dim.max == 34.5

    # Check enriched metadata
    date = dataset.metadata.date
    if date:
        assert isinstance(date, datetime)
        assert date == datetime(
            2024, 1, 1, 10, 8, 38, 979000, tzinfo=timezone.utc
        )

    game_week = dataset.metadata.game_week
    if game_week:
        assert isinstance(game_week, str)
        assert game_week == "1"

    game_id = dataset.metadata.game_id
    if game_id:
        assert isinstance(game_id, str)
        assert game_id == "288226"


def test_json_metadata(
    ball_feeds: List[Path],
    player_centroid_feeds: List[Path],
    meta_data_json: Path,
):
    dataset = hawkeye.load(
        ball_feeds=ball_feeds,
        player_centroid_feeds=player_centroid_feeds,
        meta_data=meta_data_json,
        coordinates="hawkeye",
    )
    # Check pitch dimensions
    pitch_dimensions = dataset.metadata.pitch_dimensions
    assert pitch_dimensions.x_dim.min == -52
    assert pitch_dimensions.x_dim.max == 52
    assert pitch_dimensions.y_dim.min == -33.5
    assert pitch_dimensions.y_dim.max == 33.5

    # Check enriched metadata
    date = dataset.metadata.date
    if date:
        assert isinstance(date, datetime)
        assert date == datetime(
            2024, 1, 1, 10, 8, 38, 979000, tzinfo=timezone.utc
        )

    game_week = dataset.metadata.game_week
    if game_week:
        assert isinstance(game_week, str)
        assert game_week == "1"

    game_id = dataset.metadata.game_id
    if game_id:
        assert isinstance(game_id, str)
        assert game_id == "288226"


def test_correct_normalized_deserialization(
    ball_feeds: List[Path],
    player_centroid_feeds: List[Path],
):
    dataset = hawkeye.load(
        ball_feeds=ball_feeds,
        player_centroid_feeds=player_centroid_feeds,
    )

    home_player = dataset.metadata.teams[0].players[3]
    assert dataset.records[0].players_coordinates[home_player] == Point(
        x=0.7072108495803107, y=0.5707323270685533
    )
    assert (
        dataset.records[100].players_data[home_player].speed
        == 3.963825734080879
    )

    # Check normalised pitch dimensions
    pitch_dimensions = dataset.metadata.pitch_dimensions
    assert pitch_dimensions.x_dim.min == 0.0
    assert pitch_dimensions.x_dim.max == 1.0
    assert pitch_dimensions.y_dim.min == 0.0
    assert pitch_dimensions.y_dim.max == 1.0



import json
import logging
import warnings
from datetime import datetime, timedelta, timezone
from typing import IO, Dict, NamedTuple, Optional, Union, List

from kloppy.domain import (
    AttackingDirection,
    BallState,
    DatasetFlag,
    Ground,
    Metadata,
    Orientation,
    Period,
    Player,
    PlayerData,
    Point,
    Point3D,
    PositionType,
    Provider,
    Score,
    Team,
    TrackingDataset,
    attacking_directions_from_multi_frames,
)
from kloppy.domain.services.frame_factory import create_frame
from kloppy.exceptions import DeserializationError
from kloppy.infra.serializers.tracking.deserializer import (
    TrackingDataDeserializer,
)
from kloppy.utils import performance_logging

logger = logging.getLogger(__name__)

frame_rate = 10

position_types_mapping: Dict[int, PositionType] = {
    1: PositionType.Unknown,
    2: PositionType.CenterBack,  # Provider: CB
    3: PositionType.LeftCenterBack,  # Provider: LCB
    4: PositionType.RightCenterBack,  # Provider: RCB
    5: PositionType.LeftBack,  # Provider: LWB (mapped to Left Back)
    6: PositionType.RightBack,  # Provider: RWB (mapped to Right Back)
    7: PositionType.DefensiveMidfield,  # Provider: DM
    8: PositionType.CenterMidfield,  # Provider: CM
    9: PositionType.LeftMidfield,  # Provider: LM
    10: PositionType.RightMidfield,  # Provider: RM
    11: PositionType.AttackingMidfield,  # Provider: AM
    12: PositionType.LeftWing,  # Provider: LW
    13: PositionType.RightWing,  # Provider: RW
    14: PositionType.LeftForward,  # Provider: LF
    15: PositionType.Striker,  # Provider: CF (mapped to Striker)
    16: PositionType.RightForward,  # Provider: RF
    17: PositionType.Unknown,  # Provider: SUB (mapped to Unknown)
}


class SkillCornerInputs(NamedTuple):
    meta_data: IO[bytes]
    raw_data: IO[bytes]


class SkillCornerDeserializer(TrackingDataDeserializer[SkillCornerInputs]):
    def __init__(
        self,
        limit: Optional[int] = None,
        sample_rate: Optional[float] = None,
        coordinate_system: Optional[Union[str, Provider]] = None,
        include_empty_frames: Optional[bool] = False,
        data_version: Optional[str] = None,
    ):
        super().__init__(limit, sample_rate, coordinate_system)
        self.include_empty_frames = include_empty_frames
        self.data_version = data_version

    @property
    def provider(self) -> Provider:
        return Provider.SKILLCORNER

    @classmethod
    def _get_frame_data_v2(
        cls,
        teams,
        teamdict,
        players,
        player_id_to_team_dict,
        periods,
        player_dict,
        anon_players,
        ball_id,
        referee_dict,
        frame,
    ):
        frame_id = frame["frame"]
        frame_time = cls._calc_frame_time(frame["period"], frame["time"])
        ball_owning_team = cls._get_ball_owning_team(
            frame["possession"], teams
        )

        ball_coordinates = None
        players_data = {}

        for frame_record in frame["data"]:
            # containing x, y, trackable_object, track_id, group_name
            x = frame_record.get("x")
            y = frame_record.get("y")

            trackable_object = frame_record.get("trackable_object", None)

            track_id = frame_record.get("track_id", None)
            group_name = frame_record.get("group_name", None)

            if trackable_object == ball_id or group_name == "balls":
                group_name = "balls"
                z = frame_record.get("z")
                if z is not None:
                    z = float(z)
                ball_coordinates = Point3D(x=float(x), y=float(y), z=z)
                continue

            elif (
                trackable_object in referee_dict.keys()
                or group_name == "referee"
            ):
                group_name = "referee"
                continue  # Skip Referee Coords

            if group_name is None:
                group_name = teamdict.get(
                    player_id_to_team_dict.get(trackable_object)
                )

                if group_name == "home_team":
                    player = players["HOME"][trackable_object]
                elif group_name == "away_team":
                    player = players["AWAY"][trackable_object]

            if trackable_object is None:
                player_id = str(track_id)
                if group_name == "home team":
                    if f"anon_{player_id}" not in anon_players["HOME"].keys():
                        player = cls.__create_anon_player(teams, frame_record)
                        anon_players["HOME"][f"anon_home_{player_id}"] = player
                    else:
                        player = anon_players["HOME"][f"anon_home_{player_id}"]

                elif group_name == "away team":
                    if f"anon_{player_id}" not in anon_players["AWAY"].keys():
                        player = cls.__create_anon_player(teams, frame_record)
                        anon_players["AWAY"][f"anon_away_{player_id}"] = player
                    else:
                        player = anon_players["AWAY"][f"anon_away_{player_id}"]

            players_data[player] = PlayerData(coordinates=Point(x, y))

        frame = create_frame(
            frame_id=frame_id,
            timestamp=frame_time,
            ball_coordinates=ball_coordinates,
            players_data=players_data,
            period=periods[frame["period"]],
            ball_state=BallState.ALIVE
            if ball_owning_team is not None
            else BallState.DEAD,
            ball_owning_team=ball_owning_team,
            other_data={},
        )

        return frame

    @classmethod
    def _get_frame_data_v3(
        cls,
        teams,
        teamdict,
        players,
        player_id_to_team_dict,
        periods,
        player_dict,
        anon_players,
        ball_id,
        referee_dict,
        frame,
    ):
        frame_id = frame["frame"]
        frame_time = cls._calc_frame_time(frame["period"], frame["time"])
        ball_owning_team = cls._get_ball_owning_team(
            frame["possession"], teams
        )

        ball_coordinates = cls._raw_coordinates_to_point(frame["ball_data"])

        players_data = {}

        all_players_mapping = {}
        for team in teams:
            for player in team.players:
                all_players_mapping[player.player_id] = player

        raw_players_data = frame["player_data"]
        for raw_player_data in raw_players_data:
            player = all_players_mapping[str(raw_player_data["player_id"])]
            player_coordinates = cls._raw_coordinates_to_point(raw_player_data)
            if player_coordinates:
                players_data[player] = PlayerData(
                    coordinates=player_coordinates
                )

        frame = create_frame(
            frame_id=frame_id,
            timestamp=frame_time,
            ball_coordinates=ball_coordinates,
            players_data=players_data,
            period=periods[frame["period"]],
            ball_state=BallState.ALIVE
            if ball_owning_team is not None
            else BallState.DEAD,
            ball_owning_team=ball_owning_team,
            other_data={},
        )

        return frame

    @classmethod
    def _raw_coordinates_to_point(
        cls, raw_coordinates: Dict
    ) -> Optional[Union[Point, Point3D]]:
        """
        Converts raw coordinates to a Point or Point3D object.
        """
        x = raw_coordinates["x"]
        y = raw_coordinates["y"]
        z = raw_coordinates.get("z")
        if x and y and z:
            return Point3D(x=float(x), y=float(y), z=float(z))
        elif x and y:
            return Point(x=float(x), y=float(y))

    @classmethod
    def _calc_frame_time(cls, frame_period: int, frame_time_: str):
        frame_time = cls._timestamp_from_timestring(frame_time_)

        if frame_period == 1:
            frame_time -= timedelta(seconds=0)
        elif frame_period == 2:
            frame_time -= timedelta(seconds=45 * 60)
        # TODO: check if the below is correct; just guessing here
        elif frame_period == 3:
            frame_time -= timedelta(seconds=90 * 60)
        elif frame_period == 4:
            frame_time -= timedelta(seconds=105 * 60)
        elif frame_period == 5:
            frame_time -= timedelta(seconds=120 * 60)
        else:
            raise ValueError(f"Unknown period id {frame_period}")

        return frame_time

    @classmethod
    def _get_ball_owning_team(
        cls, possession: Dict, teams: List[Team]
    ) -> Optional[Team]:
        ball_owning_team = possession.get("group")

        if ball_owning_team == "home team":
            return teams[0]
        elif ball_owning_team == "away team":
            return teams[1]

    @classmethod
    def _timestamp_from_timestring(cls, timestring):
        parts = timestring.split(":")

        if len(parts) == 2:
            m, s = parts
            return timedelta(seconds=60 * float(m) + float(s))
        elif len(parts) == 3:
            h, m, s = parts
            return timedelta(
                seconds=3600 * float(h) + 60 * float(m) + float(s)
            )
        else:
            raise ValueError("Invalid timestring format")

    @staticmethod
    def __replace_timestamp(obj):
        if "timestamp" in obj:
            obj["time"] = obj.pop("timestamp")
        return obj

    def __load_json_raw(self, file):
        # Extract the first few bytes
        start_byte = file.read(1)
        file.seek(0)

        # Check if it starts with '{' or '['
        if start_byte == b"[":
            # It's a JSON array
            try:
                data = json.load(file)
                for line in data:
                    line = self.__replace_timestamp(line)
                return data
            except json.JSONDecodeError:
                raise DeserializationError("Could not parse JSON data")

        elif start_byte == b"{":
            # It's a JSONL file
            try:
                data = []
                for line in file:
                    obj = json.loads(line)
                    data.append(self.__replace_timestamp(obj))
                return data
            except json.JSONDecodeError:
                raise DeserializationError("Could not parse JSONL data")

        raise DeserializationError("Could not determine raw data format")

    @classmethod
    def __get_periods(cls, tracking):
        """gets the Periods contained in the tracking data"""
        periods = {}

        # Extract unique periods while filtering out None values
        unique_periods = {
            frame["period"]
            for frame in tracking
            if frame["period"] is not None
        }

        for period in unique_periods:
            # Filter frames that belong to the current period and have valid "time"
            _frames = [
                frame
                for frame in tracking
                if frame["period"] == period and frame["time"] is not None
            ]

            # Ensure _frames is not empty before accessing the first and last elements
            if _frames:
                periods[period] = Period(
                    id=period,
                    start_timestamp=timedelta(
                        seconds=_frames[0]["frame"] / frame_rate
                    ),
                    end_timestamp=timedelta(
                        seconds=_frames[-1]["frame"] / frame_rate
                    ),
                )

        return periods

    @classmethod
    def __create_anon_player(cls, teams, frame_record):
        """
        creates a Player object for a track_id'ed player with known team membership but unknown identity.

        Args:
            frame_record (dict): dictionary containing 'x', 'y', 'track_id' and 'group_name'

        Returns:
            kloppy.domain.models.common.Player

        """
        track_id = frame_record.get("track_id", None)
        group_name = frame_record.get("group_name", None)

        if group_name == "home team":
            team = teams[0]
        elif group_name == "away team":
            team = teams[1]
        else:
            raise ValueError(
                f"anonymous player with track_id `{track_id}` does not have a specified group_name."
            )

        return Player(
            player_id=f"{team.ground}_anon_{track_id}",
            team=team,
            jersey_no=None,
            name=f"Anon_{track_id}",
            first_name="Anon",
            last_name=track_id,
            starting=None,
            starting_position=None,
            attributes={},
        )

    def deserialize(self, inputs: SkillCornerInputs) -> TrackingDataset:
        metadata = json.load(inputs.meta_data)
        raw_data = self.__load_json_raw(inputs.raw_data)

        with performance_logging("Loading metadata", logger=logger):
            periods = self.__get_periods(raw_data)

            teamdict = {
                metadata["home_team"].get("id"): "home_team",
                metadata["away_team"].get("id"): "away_team",
            }

            player_to_team_dict = {
                player["trackable_object"]: player["team_id"]
                for player in metadata["players"]
            }

            player_dict = {
                player["trackable_object"]: player
                for player in metadata["players"]
            }

            referee_dict = {
                ref["trackable_object"]: "referee"
                for ref in metadata["referees"]
            }
            ball_id = metadata["ball"]["trackable_object"]

            # there are different pitch_sizes in SkillCorner
            pitch_size_width = metadata["pitch_width"]
            pitch_size_length = metadata["pitch_length"]

            transformer = self.get_transformer(
                pitch_length=pitch_size_length, pitch_width=pitch_size_width
            )

            home_team_id = metadata["home_team"]["id"]
            away_team_id = metadata["away_team"]["id"]

            players = {"HOME": {}, "AWAY": {}}

            home_team = Team(
                team_id=home_team_id,
                name=metadata["home_team"]["name"],
                ground=Ground.HOME,
            )
            away_team = Team(
                team_id=away_team_id,
                name=metadata["away_team"]["name"],
                ground=Ground.AWAY,
            )
            teams = [home_team, away_team]

            date = metadata.get("date_time")
            if date:
                date = datetime.strptime(date, "%Y-%m-%dT%H:%M:%SZ").replace(
                    tzinfo=timezone.utc
                )

            game_id = metadata.get("id")
            if game_id:
                game_id = str(game_id)

            home_team_coach = metadata.get("home_team_coach")
            home_coach = (
                f"{home_team_coach['first_name']} {home_team_coach['last_name']}"
                if home_team_coach is not None
                else None
            )

            away_team_coach = metadata.get("away_team_coach")
            away_coach = (
                f"{away_team_coach['first_name']} {away_team_coach['last_name']}"
                if away_team_coach is not None
                else None
            )

            if game_id:
                game_id = str(game_id)

            for player_track_obj_id, player in player_dict.items():
                team_id = player["team_id"]

                if team_id == home_team_id:
                    team_string = "HOME"
                    team = home_team
                elif team_id == away_team_id:
                    team_string = "AWAY"
                    team = away_team

                players[team_string][player_track_obj_id] = Player(
                    player_id=f"{player['id']}",
                    team=team,
                    jersey_no=player["number"],
                    name=f"{player['first_name']} {player['last_name']}",
                    first_name=player["first_name"],
                    last_name=player["last_name"],
                    starting=player["start_time"] == "00:00:00",
                    starting_position=position_types_mapping.get(
                        player["player_role"]["id"], PositionType.Unknown
                    ),
                    attributes={},
                )

            home_team.players = list(players["HOME"].values())
            away_team.players = list(players["AWAY"].values())

        anon_players = {"HOME": {}, "AWAY": {}}

        with performance_logging("Loading data", logger=logger):

            def _iter():
                n = 0
                sample = 1.0 / self.sample_rate

                for frame in raw_data:
                    frame_period = frame["period"]

                    if frame_period is not None:
                        if n % sample == 0:
                            yield frame
                        n += 1

        frames = []

        get_frame_data = (
            self._get_frame_data_v2
            if self.data_version == "V2"
            else self._get_frame_data_v3
        )

        n_frames = 0
        for _frame in _iter():
            # include frame if there is any tracking data, players or ball.
            # or if include_empty_frames == True
            not_empty_frame = (
                len(_frame["data"]) > 0
                if self.data_version == "V2"
                else len(_frame["player_data"]) > 0
            )
            if self.include_empty_frames or not_empty_frame:
                frame = get_frame_data(
                    teams,
                    teamdict,
                    players,
                    player_to_team_dict,
                    periods,
                    player_dict,
                    anon_players,
                    ball_id,
                    referee_dict,
                    _frame,
                )

                frame = transformer.transform_frame(frame)

                frames.append(frame)
                n_frames += 1

                if self.limit and n_frames + 1 >= (
                    self.limit / self.sample_rate
                ):
                    break

        attacking_directions = attacking_directions_from_multi_frames(
            frames, list(periods.values())
        )
        if attacking_directions[1] == AttackingDirection.LTR:
            orientation = Orientation.HOME_AWAY
        elif attacking_directions[1] == AttackingDirection.RTL:
            orientation = Orientation.AWAY_HOME
        else:
            warnings.warn(
                "Could not determine orientation of dataset, defaulting to NOT_SET"
            )
            orientation = Orientation.NOT_SET

        metadata = Metadata(
            teams=teams,
            periods=sorted(periods.values(), key=lambda p: p.id),
            pitch_dimensions=transformer.get_to_coordinate_system().pitch_dimensions,
            score=Score(
                home=metadata["home_team_score"],
                away=metadata["away_team_score"],
            ),
            frame_rate=10,
            orientation=orientation,
            provider=Provider.SKILLCORNER,
            flags=DatasetFlag.BALL_STATE | DatasetFlag.BALL_OWNING_TEAM,
            coordinate_system=transformer.get_to_coordinate_system(),
            date=date,
            game_id=game_id,
            home_coach=home_coach,
            away_coach=away_coach,
        )

        return TrackingDataset(
            records=frames,
            metadata=metadata,
        )



import json
from typing import Type, Union

from kloppy.config import get_config
from kloppy.domain import EventDataset, EventFactory, List, Optional
from kloppy.infra.serializers.event.wyscout import (
    WyscoutDeserializerV2,
    WyscoutDeserializerV3,
    WyscoutInputs,
)
from kloppy.io import FileLike, open_as_file


def load(
    event_data: FileLike,
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
    data_version: Optional[str] = None,
) -> EventDataset:
    """
    Load Wyscout event data.

    Args:
        event_data: JSON feed with the raw event data of a game.
        event_types: A list of event types to load.
        coordinates: The coordinate system to use.
        event_factory: A custom event factory.
        data_version: The version of the Wyscout data. Supported versions are "V2" and "V3".

    Returns:
        The parsed event data.
    """
    if data_version == "V2":
        deserializer_class = WyscoutDeserializerV2
    elif data_version == "V3":
        deserializer_class = WyscoutDeserializerV3
    else:
        deserializer_class = identify_deserializer(event_data)

    deserializer = deserializer_class(
        event_types=event_types,
        coordinate_system=coordinates,
        event_factory=event_factory or get_config("event_factory"),
    )

    with open_as_file(event_data) as event_data_fp:
        return deserializer.deserialize(
            inputs=WyscoutInputs(event_data=event_data_fp),
        )


def load_open_data(
    match_id: Union[str, int] = "2499841",
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    """
    Load Wyscout open data.

    This dataset is a public release of event stream data, collected by Wyscout
    containing all matches of the 2017/18 season of the top-5 European leagues
    (La Liga, Serie A, Bundesliga, Premier League, Ligue 1), the FIFA World
    Cup 2018, and UEFA Euro Cup 2016. For a detailed description,
    see Pappalardo et al. [1].

    Args:
        match_id: The id of the match to load data for.
        event_types: A list of event types to load.
        coordinates: The coordinate system to use.
        event_factory: A custom event factory.

    Returns:
        The parsed event data.

    References:
        [1] Pappalardo, L., Cintia, P., Rossi, A. et al. A public data set of spatio-temporal match events in soccer competitions. Sci Data 6, 236 (2019). https://doi.org/10.1038/s41597-019-0247-7
    """
    return load(
        event_data=f"https://raw.githubusercontent.com/koenvo/wyscout-soccer-match-event-dataset/main/processed-v2/files/{match_id}.json",
        event_types=event_types,
        coordinates=coordinates,
        event_factory=event_factory,
    )


def identify_deserializer(
    event_data: FileLike,
) -> Union[Type[WyscoutDeserializerV3], Type[WyscoutDeserializerV2]]:
    with open_as_file(event_data) as event_data_fp:
        events_with_meta = json.load(event_data_fp)

    events = events_with_meta["events"]
    first_event = events[0]

    deserializer = None
    if "eventName" in first_event:
        deserializer = WyscoutDeserializerV2
    elif "primary" in first_event.get("type", {}):
        deserializer = WyscoutDeserializerV3

    if deserializer is None:
        raise ValueError(
            "Wyscout data version could not be recognized, please specify"
        )

    return deserializer



import json
import logging
from datetime import datetime, timedelta, timezone
from dateutil.parser import parse
import warnings
import re
from lxml import objectify

from typing import (
    IO,
    Any,
    Dict,
    List,
    NamedTuple,
    Optional,
    Union,
    Iterable,
    Callable,
)
from itertools import zip_longest

from kloppy.domain import (
    AttackingDirection,
    Frame,
    Ground,
    Metadata,
    Orientation,
    Period,
    Player,
    PlayerData,
    Point,
    Point3D,
    Provider,
    Team,
    TrackingDataset,
    attacking_direction_from_frame,
)
from kloppy.utils import performance_logging
from kloppy.io import FileLike, open_as_file, get_file_extension
from kloppy.exceptions import DeserializationError

from ..deserializer import TrackingDataDeserializer

try:
    import tqdm
except ImportError:
    tqdm = None

logger = logging.getLogger(__name__)


class HawkEyeInputs(NamedTuple):
    ball_feeds: Iterable[FileLike]
    player_centroid_feeds: Iterable[FileLike]
    meta_data: Optional[FileLike] = None
    show_progress: Optional[bool] = False


class HawkEyeObjectIdentifier:
    FIFA_ID = "fifaId"
    UEFA_ID = "uefaId"
    HE_ID = "heId"

    PRIORITY_IDS = [FIFA_ID, UEFA_ID, HE_ID]

    @classmethod
    def get_identifier_variable(cls, player_tracking_data):
        player_object = player_tracking_data["details"]["players"][0]["id"]

        object_id = cls.HE_ID
        for identifier in cls.PRIORITY_IDS:
            if player_object.get(identifier):
                return identifier
        return object_id


class HawkEyeDeserializer(TrackingDataDeserializer[HawkEyeInputs]):
    def __init__(
        self,
        pitch_width: Optional[float] = 68.0,
        pitch_length: Optional[float] = 105.0,
        limit: Optional[int] = None,
        sample_rate: Optional[float] = None,
        coordinate_system: Optional[Union[str, Provider]] = None,
    ):
        super().__init__(limit, sample_rate, coordinate_system)
        self.object_id: HawkEyeObjectIdentifier = None
        self.pitch_width = pitch_width
        self.pitch_length = pitch_length

        self._game_id = None
        self._game_week = None
        self._game_date = None

    @property
    def provider(self) -> Provider:
        return Provider.HAWKEYE

    @staticmethod
    def __parse_periods(
        raw_periods: List[Dict[str, Any]]
    ) -> Dict[str, Period]:
        parsed_periods = {}
        for period in raw_periods:
            if period["segment"] == 0:
                continue
            period_id = period["segment"]

            parsed_periods[period_id] = Period(
                id=period_id,
                start_timestamp=datetime.strptime(
                    period["startTimeUTC"], "%Y-%m-%dT%H:%M:%S.%fZ"
                ),
                end_timestamp=datetime.strptime(
                    period["endTimeUTC"], "%Y-%m-%dT%H:%M:%S.%fZ"
                ),
            )
        return parsed_periods

    def __parse_teams(
        self, raw_teams: List[Dict[str, Any]]
    ) -> Dict[str, Team]:
        parsed_teams = {}
        for team in raw_teams:
            team_id = team["id"][self.object_id]
            parsed_teams[team_id] = Team(
                team_id=team["id"][self.object_id],
                name=team["name"],
                ground=Ground.HOME if team["home"] else Ground.AWAY,
            )
        return parsed_teams

    def __parse_players(
        self, raw_players: List[Dict[str, Any]], teams: Dict[str, Team]
    ) -> Dict[str, Player]:
        parsed_players = {}
        for player in raw_players:
            player_id = player["id"][self.object_id]
            parsed_players[player_id] = Player(
                player_id=player["id"][self.object_id],
                team=teams[player["teamId"][self.object_id]],
                jersey_no=int(player["jerseyNumber"]),
                starting_position=player["role"]["name"],
            )
        return parsed_players

    @staticmethod
    def __infer_frame_rate(ball_tracking_data, n_samples=25):
        total_time_difference = 0

        for i in range(len(ball_tracking_data["samples"]["ball"]) - 1):
            current_time = ball_tracking_data["samples"]["ball"][i]["time"]
            next_time = ball_tracking_data["samples"]["ball"][i + 1]["time"]

            time_difference = next_time - current_time
            total_time_difference += time_difference

            if i >= n_samples - 1:
                break

        return int(1 / (total_time_difference / n_samples))

    def __parse_meta_data_json(self, meta_data):
        with open_as_file(meta_data) as meta_data_fp:
            meta_data = json.load(meta_data_fp)

        kick_off_time = meta_data.get("KickOffTime")
        match_day = re.search(r"\d+", meta_data.get("MatchDay"))
        game_week = match_day.group() if match_day else None

        stadium = meta_data.get("Stadium")
        self.pitch_length = (
            stadium.get("PitchLength", self.pitch_length)
            if stadium
            else self.pitch_length
        )
        self.pitch_width = (
            stadium.get("PitchWidth", self.pitch_width)
            if stadium
            else self.pitch_width
        )

        self._game_id = meta_data.get("MatchId")
        self._game_date = (
            parse(kick_off_time.get("DateTime")).replace(tzinfo=timezone.utc)
            if kick_off_time
            else None
        )
        self._game_week = game_week

    def __parse_meta_data_xml(self, meta_data):
        with open_as_file(meta_data) as meta_data_fp:
            root = objectify.fromstring(meta_data_fp.read())

        game_id = getattr(root, "id", None)
        date = (
            getattr(root.kickOffTime, "dateTime", None)
            if hasattr(root, "kickOffTime")
            else None
        )
        game_week_attr = (
            getattr(root.matchday, "name", None)
            if hasattr(root, "matchday")
            else None
        )
        self.pitch_length = (
            getattr(root.stadium.pitch, "length", self.pitch_length)
            if hasattr(root.stadium, "pitch")
            else self.pitch_length
        )
        self.pitch_width = (
            getattr(root.stadium.pitch, "width", self.pitch_width)
            if hasattr(root.stadium, "pitch")
            else self.pitch_width
        )

        if date:
            date = parse(str(date)).replace(tzinfo=timezone.utc)
        if game_week_attr:
            match_day = re.search(r"\d+", str(game_week_attr))
            game_week = match_day.group() if match_day else None

        self._game_id = str(game_id) if game_id else None
        self._game_date = date
        self._game_week = game_week

    def __parse_meta_data(self, meta_data):
        if meta_data is None:
            return

        meta_data_extension = get_file_extension(meta_data)
        if meta_data_extension == ".xml":
            return self.__parse_meta_data_xml(meta_data)
        elif meta_data_extension == ".json":
            return self.__parse_meta_data_json(meta_data)
        else:
            raise ValueError(
                "Metadata only supports .json and .xml file formats..."
            )

    def deserialize(self, inputs: HawkEyeInputs) -> TrackingDataset:

        self.__parse_meta_data(inputs.meta_data)

        transformer = self.get_transformer(
            pitch_length=self.pitch_length,
            pitch_width=self.pitch_width,
        )

        parsed_teams = {}
        parsed_players = {}
        parsed_periods = {}
        parsed_periods = {}
        parsed_frames = {}
        frame_rate = None

        it = list(zip_longest(inputs.ball_feeds, inputs.player_centroid_feeds))
        if inputs.show_progress:
            if tqdm is None:
                warnings.warn(
                    "tqdm not installed, progress bar will not be shown"
                )
            else:
                it = tqdm.tqdm(it)

        for ball_feed, player_centroid_feed in it:
            # Read the ball and player tracking data feeds.
            with open_as_file(ball_feed) as ball_data_fp:
                ball_tracking_data = json.load(ball_data_fp)
            with open_as_file(player_centroid_feed) as player_centroid_data_fp:
                player_tracking_data = json.load(player_centroid_data_fp)

            self.object_id = HawkEyeObjectIdentifier.get_identifier_variable(
                player_tracking_data
            )

            if frame_rate is None:
                frame_rate = self.__infer_frame_rate(ball_tracking_data)

            if not self._game_id:
                self._game_id = ball_tracking_data["details"]["match"]["id"][
                    self.object_id
                ]

            # Parse the teams, players and periods. A value can be added by
            # later feeds, but we will not overwrite existing values.
            with performance_logging("Parsing meta data", logger=logger):
                parsed_teams = {
                    **self.__parse_teams(
                        ball_tracking_data["details"]["teams"]
                    ),
                    **parsed_teams,
                }
                parsed_players = {
                    **self.__parse_players(
                        player_tracking_data["details"]["players"],
                        parsed_teams,
                    ),
                    **parsed_players,
                }
                parsed_periods = {
                    **self.__parse_periods(ball_tracking_data["segments"]),
                    **parsed_periods,
                }

            # Parse the ball tracking data
            period_id = ball_tracking_data["sequences"]["segment"]
            minute = ball_tracking_data["sequences"]["match-minute"] - 1

            period_minute = (
                minute
                if period_id == 1
                else (minute - 45)
                if period_id == 2
                else (minute - 90)
                if period_id == 3
                else (minute - 105)
                if period_id == 4
                else (minute - 120)
            )

            with performance_logging(
                "Parsing ball tracking data", logger=logger
            ):
                for detection in ball_tracking_data["samples"]["ball"]:
                    frame_id = int(
                        (minute * 60 + float(detection["time"])) * frame_rate
                    )
                    parsed_frames[frame_id] = Frame(
                        frame_id=frame_id,
                        timestamp=timedelta(
                            minutes=period_minute, seconds=detection["time"]
                        ),
                        ball_coordinates=Point3D(
                            x=detection["pos"][0],
                            y=detection["pos"][1],
                            z=detection["pos"][2],
                        ),
                        ball_speed=detection["speed"]["mps"],
                        ball_state=None,
                        ball_owning_team=None,
                        players_data={},
                        period=parsed_periods[period_id],
                        other_data={},
                        statistics=[],
                    )

            # Parse the player tracking data
            _period_id = player_tracking_data["sequences"]["segment"]
            _minute = player_tracking_data["sequences"]["match-minute"] - 1

            if _period_id != period_id or _minute != minute:
                raise DeserializationError(
                    "The feed for ball tracking and player tracking are not in sync"
                )
            with performance_logging(
                "Parsing player tracking data", logger=logger
            ):
                for detection in player_tracking_data["samples"]["people"]:
                    if detection["role"]["name"] not in [
                        "Outfielder",
                        "Goalkeeper",
                    ]:
                        continue
                    player = parsed_players[
                        detection["personId"][self.object_id]
                    ]
                    for centroid in detection["centroid"]:
                        frame_id = int(
                            (minute * 60 + centroid["time"]) * frame_rate
                        )
                        player_data = PlayerData(
                            coordinates=Point(
                                x=centroid["pos"][0],
                                y=centroid["pos"][1],
                            ),
                            distance=centroid["distance"]["metres"],
                            speed=centroid["speed"]["mps"],
                        )
                        if frame_id in parsed_frames:
                            parsed_frames[frame_id].players_data[
                                player
                            ] = player_data
                        else:
                            parsed_frames[frame_id] = Frame(
                                frame_id=frame_id,
                                timestamp=timedelta(
                                    minutes=period_minute,
                                    seconds=centroid["time"],
                                ),
                                ball_coordinates=Point3D(
                                    float("nan"), float("nan"), float("nan")
                                ),
                                ball_state=None,
                                ball_owning_team=None,
                                players_data={player: player_data},
                                period=parsed_periods[period_id],
                                other_data={},
                                statistics=[],
                            )

            if self.limit and len(parsed_frames) >= self.limit:
                break

        # Convert the parsed frames to a list
        frame_ts = sorted(parsed_frames.keys())
        frames = []
        sample = 1.0 / self.sample_rate
        for i, ts in enumerate(frame_ts):
            if ts % sample == 0:
                frame = transformer.transform_frame(parsed_frames[ts])
                frames.append(frame)

                if self.limit and i * (self.sample_rate) + 1 >= self.limit:
                    break

        # Add player list to teams
        for team in parsed_teams.values():
            for player in parsed_players.values():
                if player.team == team:
                    team.players.append(player)

        try:
            first_frame = next(
                frame for frame in frames if frame.period.id == 1
            )
            orientation = (
                Orientation.HOME_AWAY
                if attacking_direction_from_frame(first_frame)
                == AttackingDirection.LTR
                else Orientation.AWAY_HOME
            )
        except StopIteration:
            warnings.warn(
                "Could not determine orientation of dataset, defaulting to NOT_SET"
            )
            orientation = Orientation.NOT_SET

        meta_data = Metadata(
            teams=list(parsed_teams.values()),
            periods=list(parsed_periods.values()),
            pitch_dimensions=transformer.get_to_coordinate_system().pitch_dimensions,
            score=None,
            frame_rate=frame_rate,
            orientation=orientation,
            provider=Provider.HAWKEYE,
            flags=None,
            coordinate_system=transformer.get_to_coordinate_system(),
            game_id=self._game_id,
            date=self._game_date,
            game_week=self._game_week,
        )

        return TrackingDataset(
            records=frames,
            metadata=meta_data,
        )



import warnings
from typing import Union

from kloppy.config import get_config
from kloppy.domain import EventDataset, EventFactory, List, Optional
from kloppy.domain.models.statsbomb.event import StatsBombEventFactory
from kloppy.infra.serializers.event.statsbomb import (
    StatsBombDeserializer,
    StatsBombInputs,
)
from kloppy.io import FileLike, Source, open_as_file


def load(
    event_data: FileLike,
    lineup_data: FileLike,
    three_sixty_data: Optional[FileLike] = None,
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
    additional_metadata: dict = {},
) -> EventDataset:
    """
    Load StatsBomb event data.

    Args:
        event_data: JSON feed with the raw event data of a game.
        lineup_data: JSON feed with the corresponding lineup information of the game.
        three_sixty_data: JSON feed with the 360 freeze frame data of the game.
        event_types: A list of event types to load.
        coordinates: The coordinate system to use.
        event_factory: A custom event factory.
        additional_metadata: A dict with additional data that will be added to
            the metadata. See the [`Metadata`][kloppy.domain.Metadata] entity
            for a list of possible keys.

    Returns:
        The parsed event data.
    """
    deserializer = StatsBombDeserializer(
        event_types=event_types,
        coordinate_system=coordinates,
        event_factory=event_factory
        or get_config("event_factory")
        or StatsBombEventFactory(),
    )
    with open_as_file(event_data) as event_data_fp, open_as_file(
        lineup_data
    ) as lineup_data_fp, open_as_file(
        Source.create(three_sixty_data, optional=True)
    ) as three_sixty_data_fp:
        return deserializer.deserialize(
            inputs=StatsBombInputs(
                event_data=event_data_fp,
                lineup_data=lineup_data_fp,
                three_sixty_data=three_sixty_data_fp,
            ),
            additional_metadata=additional_metadata,
        )


def load_open_data(
    match_id: Union[str, int] = "15946",
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
    event_factory: Optional[EventFactory] = None,
) -> EventDataset:
    """
    Load StatsBomb open data.

    This function loads event data directly from the StatsBomb open data
    GitHub repository.

    Args:
        match_id: The id of the match to load data for.
        event_types: A list of event types to load.
        coordinates: The coordinate system to use.
        event_factory: A custom event factory.

    Returns:
        The parsed event data.
    """
    warnings.warn(
        "\n\nYou are about to use StatsBomb public data."
        "\nBy using this data, you are agreeing to the user agreement. "
        "\nThe user agreement can be found here: https://github.com/statsbomb/open-data/blob/master/LICENSE.pdf"
        "\n"
    )

    return load(
        event_data=f"https://raw.githubusercontent.com/statsbomb/open-data/master/data/events/{match_id}.json",
        lineup_data=f"https://raw.githubusercontent.com/statsbomb/open-data/master/data/lineups/{match_id}.json",
        three_sixty_data=Source(
            f"https://raw.githubusercontent.com/statsbomb/open-data/master/data/three-sixty/{match_id}.json",
            skip_if_missing=True,
        ),
        event_types=event_types,
        coordinates=coordinates,
        event_factory=event_factory,
    )



"""I/O utilities for reading raw data."""

import bz2
import contextlib
import gzip
import logging
import lzma
import os
import re
from dataclasses import dataclass, replace
from io import BufferedWriter, BytesIO, TextIOWrapper
from typing import (
    IO,
    Any,
    BinaryIO,
    Callable,
    ContextManager,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

from kloppy.exceptions import AdapterError, InputNotFoundError
from kloppy.infra.io.adapters import get_adapter

logger = logging.getLogger(__name__)

DEFAULT_GZIP_COMPRESSION = 1
DEFAULT_BZ2_COMPRESSION = 9
DEFAULT_XZ_COMPRESSION = 6


FilePath = Union[str, bytes, os.PathLike]
FileOrPath = Union[FilePath, IO]


@dataclass(frozen=True)
class Source:
    """A wrapper around a file-like object to enable optional inputs.

    Args:
        data (FileLike): The file-like object.
        optional (bool): Whether the file is optional. Defaults to False.
        skip_if_missing (bool): Whether to skip the file if it is missing. Defaults to False.

    Example:

        >>> open_as_file(Source.create("example.csv", optional=True))
    """

    data: Optional[FileOrPath]
    optional: bool = False
    skip_if_missing: bool = False

    @classmethod
    def create(cls, input_: Optional[FileOrPath], **kwargs):
        if isinstance(input_, Source):
            return replace(input_, **kwargs)
        return Source(data=input_, **kwargs)


FileLike = Union[FileOrPath, Source]


def _file_or_path_to_binary_stream(
    file_or_path: FileOrPath, binary_mode: str
) -> Tuple[BinaryIO, bool]:
    """
    Converts a file path or a file-like object to a binary stream.

    Args:
        file_or_path: The file path or file-like object to convert.
        binary_mode: The binary mode to open the file in. Must be one of 'rb', 'wb', or 'ab'.

    Returns:
        A tuple containing the binary stream and a boolean indicating whether
        a new file was opened (True) or an existing file-like object was used (False).
    """
    assert binary_mode in ("rb", "wb", "ab")

    if isinstance(file_or_path, (str, bytes)) or hasattr(
        file_or_path, "__fspath__"
    ):
        # If file_or_path is a path-like object, open it and return the binary stream
        return open(os.fspath(file_or_path), binary_mode), True  # type: ignore

    if isinstance(file_or_path, TextIOWrapper):
        # If file_or_path is a TextIOWrapper, return its underlying binary buffer
        return file_or_path.buffer, False

    if hasattr(file_or_path, "readinto") or hasattr(file_or_path, "write"):
        # If file_or_path is a file-like object, return it as is
        return file_or_path, False  # type: ignore

    raise TypeError(
        f"Unsupported type for {file_or_path}, {file_or_path.__class__.__name__}."
    )


def _detect_format_from_content(file_or_path: FileOrPath) -> Optional[str]:
    """
    Attempts to detect file format from the content by reading the first
    6 bytes. Returns None if no format could be detected.
    """
    fileobj, closefd = _file_or_path_to_binary_stream(file_or_path, "rb")
    try:
        if not fileobj.readable():
            return None
        if hasattr(fileobj, "peek"):
            bs = fileobj.peek(6)
        elif hasattr(fileobj, "seekable") and fileobj.seekable():
            current_pos = fileobj.tell()
            bs = fileobj.read(6)
            fileobj.seek(current_pos)
        else:
            return None

        if bs[:2] == b"\x1f\x8b":
            # https://tools.ietf.org/html/rfc1952#page-6
            return "gz"
        elif bs[:3] == b"\x42\x5a\x68":
            # https://en.wikipedia.org/wiki/List_of_file_signatures
            return "bz2"
        elif bs[:6] == b"\xfd\x37\x7a\x58\x5a\x00":
            # https://tukaani.org/xz/xz-file-format.txt
            return "xz"
        return None
    finally:
        if closefd:
            fileobj.close()


def _detect_format_from_extension(filename: FilePath) -> Optional[str]:
    """
    Attempt to detect file format from the filename extension.
    Return None if no format could be detected.
    """
    extensions = ("bz2", "xz", "gz")

    if isinstance(filename, bytes):
        for ext in extensions:
            if filename.endswith(b"." + ext.encode()):
                return ext

    if isinstance(filename, str):
        for ext in extensions:
            if filename.endswith("." + ext):
                return ext

    if hasattr(filename, "name"):
        return _detect_format_from_extension(filename.name)

    return None


def _filepath_from_path_or_filelike(file_or_path: FileOrPath) -> str:
    try:
        return os.fspath(file_or_path)  # type: ignore
    except TypeError:
        pass

    if hasattr(file_or_path, "name"):
        name = file_or_path.name
        if isinstance(name, str):
            return name
        elif isinstance(name, bytes):
            return name.decode()

    return ""


def _open(
    filename: FileOrPath,
    mode: str = "rb",
    compresslevel: Optional[int] = None,
    format: Optional[str] = None,
) -> BinaryIO:
    """
        A replacement for the "open" function that can also read and write
        compressed files transparently. The supported compression formats are gzip,
        bzip2 and xz. Filename can be a string, a Path or a file object.

        When writing, the file format is chosen based on the file name extension:
        - .gz uses gzip compression
        - .bz2 uses bzip2 compression
    - .xz uses xz/lzma compression
        - otherwise, no compression is used

        When reading, if a file name extension is available, the format is detected
        using it, but if not, the format is detected from the contents.

        mode can be: 'rb', 'ab', or 'wb'.

        compresslevel is the compression level for writing to gzip, and xz.
        This parameter is ignored for the other compression formats.
        If set to None, a default depending on the format is used:
        gzip: 6, xz: 6.

        format overrides the autodetection of input and output formats. This can be
        useful when compressed output needs to be written to a file without an
        extension. Possible values are "gz", "xz", "bz2" and "raw". In case of
        "raw", no compression is used.
    """
    if mode not in ("rb", "wb", "ab"):
        raise ValueError(f"Mode '{mode}' not supported")
    filepath = _filepath_from_path_or_filelike(filename)

    if format not in (None, "gz", "xz", "bz2", "raw"):
        raise ValueError(
            f"Format not supported: {format}. Choose one of: 'gz', 'xz', 'bz2'"
        )

    if format == "raw":
        detected_format = None
    else:
        detected_format = format or _detect_format_from_extension(filepath)
        if detected_format is None and "r" in mode:
            detected_format = _detect_format_from_content(filename)

    if detected_format == "gz":
        opened_file = _open_gz(filename, mode, compresslevel)
    elif detected_format == "xz":
        opened_file = _open_xz(filename, mode, compresslevel)
    elif detected_format == "bz2":
        opened_file = _open_bz2(filename, mode, compresslevel)
    else:
        opened_file, _ = _file_or_path_to_binary_stream(filename, mode)

    return opened_file


def _open_bz2(
    filename: FileOrPath,
    mode: str,
    compresslevel: Optional[int] = None,
) -> BinaryIO:
    assert mode in ("rb", "ab", "wb")
    if compresslevel is None:
        compresslevel = DEFAULT_BZ2_COMPRESSION

    if "r" in mode:
        return bz2.open(filename, mode)  # type: ignore
    return BufferedWriter(bz2.open(filename, mode, compresslevel))  # type: ignore


def _open_xz(
    filename: FileOrPath,
    mode: str,
    compresslevel: Optional[int] = None,
) -> BinaryIO:
    assert mode in ("rb", "ab", "wb")
    if compresslevel is None:
        compresslevel = DEFAULT_XZ_COMPRESSION

    if "r" in mode:
        return lzma.open(filename, mode)  # type: ignore
    return BufferedWriter(lzma.open(filename, mode, preset=compresslevel))  # type: ignore


def _open_gz(
    filename: FileOrPath,
    mode: str,
    compresslevel: Optional[int] = None,
) -> BinaryIO:
    assert mode in ("rb", "ab", "wb")
    if compresslevel is None:
        compresslevel = DEFAULT_GZIP_COMPRESSION

    if "r" in mode:
        return gzip.open(filename, mode)  # type: ignore
    return BufferedWriter(gzip.open(filename, mode, compresslevel=compresslevel))  # type: ignore


def get_file_extension(file_or_path: FileLike) -> str:
    """Determine the file extension of the given file-like object.

    If the file has compression extensions such as '.gz', '.xz', or '.bz2',
    they will be stripped before determining the extension.

    Args:
        file_or_path (FileLike): The file-like object whose extension needs to be determined.

    Returns:
        str: The file extension, including the dot ('.') if present.

    Raises:
        Exception: If the extension cannot be determined.

    Example:

        >>> get_file_extension("example.xml.gz")
        '.xml'
        >>> get_file_extension(Path("example.txt"))
        '.txt'
        >>> get_file_extension(Source(data="example.csv"))
        '.csv'
    """
    if isinstance(file_or_path, (str, bytes)) or hasattr(
        file_or_path, "__fspath__"
    ):
        path = os.fspath(file_or_path)  # type: ignore
        for ext in [".gz", ".xz", ".bz2"]:
            if path.endswith(ext):
                path = path[: -len(ext)]
        return os.path.splitext(path)[1]

    if isinstance(file_or_path, Source):
        return get_file_extension(file_or_path.data)

    raise TypeError(
        f"Could not determine extension for input type: {type(file_or_path)}"
    )


@contextlib.contextmanager
def dummy_context_mgr() -> Generator[None, None, None]:
    yield


def open_as_file(input_: FileLike) -> ContextManager[Optional[BinaryIO]]:
    """Open a byte stream to the given input object.

    The following input types are supported:
        - A string or `pathlib.Path` object representing a local file path.
        - A string representing a URL. It should start with 'http://' or
          'https://'.
        - A string representing a path to a file in a Amazon S3 cloud storage
          bucket. It should start with 's3://'.
        - A xml or json string containing the data. The string should contain
          a '{' or '<' character. Otherwise, it will be treated as a file path.
        - A bytes object containing the data.
        - A buffered binary stream that inherits from `io.BufferedIOBase`.
        - A [Source](`kloppy.io.Source`) object that wraps any of the above
          input types.

    Args:
        input_ (FileLike): The input object to be opened.

    Returns:
        BinaryIO: A binary stream to the input object.

    Raises:
        ValueError: If the input is required but not provided.
        InputNotFoundError: If the input file is not found and should not be skipped.
        TypeError: If the input type is not supported.

    Example:

        >>> with open_as_file("example.txt") as f:
        ...     contents = f.read()

    Note:
        To support reading data from other sources, see the
        [Adapter](`kloppy.io.adapters.Adapter`) class.

        If the given file path or URL ends with '.gz', '.xz', or '.bz2', the
        file will be decompressed before being read.
    """
    if isinstance(input_, Source):
        if input_.data is None and input_.optional:
            # This saves us some additional code in every vendor specific code
            return dummy_context_mgr()
        elif input_.data is None:
            raise ValueError("Input required but not provided.")
        else:
            try:
                return open_as_file(input_.data)
            except InputNotFoundError as exc:
                if input_.skip_if_missing:
                    logging.info(f"Input {input_.data} not found. Skipping")
                    return dummy_context_mgr()
                else:
                    raise exc

    if isinstance(input_, str) and ("{" in input_ or "<" in input_):
        # If input_ is a JSON or XML string, return it as a binary stream
        return BytesIO(input_.encode("utf8"))

    if isinstance(input_, bytes):
        # If input_ is a bytes object, return it as a binary stream
        return BytesIO(input_)

    if isinstance(input_, str) or hasattr(input_, "__fspath__"):
        # If input_ is a path-like object, open it and return the binary stream
        uri = _filepath_from_path_or_filelike(input_)

        adapter = get_adapter(uri)
        if adapter:
            stream = BytesIO()
            adapter.read_to_stream(uri, stream)
            stream.seek(0)
        else:
            raise AdapterError(f"No adapter found for {uri}")
        return stream

    if isinstance(input_, TextIOWrapper):
        # If file_or_path is a TextIOWrapper, return its underlying binary buffer
        return input_.buffer

    if hasattr(input_, "readinto"):
        # If file_or_path is a file-like object, return it as is
        return _open(input_)  # type: ignore

    raise TypeError(f"Unsupported input type: {type(input_)}")


def _natural_sort_key(path: str) -> List[Union[int, str]]:
    # Split string into list of chunks for natural sorting
    return [
        int(text) if text.isdigit() else text.lower()
        for text in re.split(r"(\d+)", path)
    ]


def expand_inputs(
    inputs: Union[FileLike, Iterable[FileLike]],
    regex_filter: Optional[str] = None,
    sort_key: Optional[Callable[[str], Any]] = None,
) -> Iterator[FileLike]:
    """
    Process input data to a list of file-like objects.

    This function resolves the input data to a list of file-like objects
    that can be opened with [`open_as_file`][kloppy.io.open_as_file].
    The input data can be any of the types supported by `open_as_file`,
    including a list of these input types or a directory.

    If the input is a directory, the directory is expanded to a list of files,
    which are then filtered using a regex pattern and sorted using a custom key.

    If the input is a list of file-like objects, these are assumed to be in the
    correct order and are returned as is.

    Args:
        inputs: The input object, a list of input objects, or a directory with input objects to be opened.
        regex_filter: A regex pattern to filter files (only applies to directories).
        sort_key: A callable to define sorting logic for files (only applies to directories).
          If not provided, files are sorted as ["file1.txt", "file2.txt", "file10.txt"].

    Returns:
        An iterator over the resolved file paths or stream content.
    """

    def is_file(uri):
        adapter = get_adapter(uri)
        if adapter:
            return adapter.is_file(uri)
        raise AdapterError(f"No adapter found for {uri}")

    def is_directory(uri):
        adapter = get_adapter(uri)
        if adapter:
            return adapter.is_directory(uri)
        raise AdapterError(f"No adapter found for {uri}")

    def process_expansion(files):
        """
        Process a list of files by filtering and sorting them.

        Args:
            files: List of file URIs to process.

        Returns:
            A sorted and filtered list of file URIs.
        """
        files = [f for f in files if not is_directory(f)]

        if regex_filter:
            pattern = re.compile(regex_filter)
            files = [f for f in files if pattern.search(f)]

        files.sort(key=sort_key or _natural_sort_key)
        return files

    if isinstance(inputs, (str, os.PathLike)):
        uri = _filepath_from_path_or_filelike(inputs)

        if is_directory(uri):
            adapter = get_adapter(uri)
            if adapter:
                yield from process_expansion(
                    adapter.list_directory(uri, recursive=True)
                )
            else:
                raise AdapterError(f"No adapter found for {uri}")
        elif is_file(uri):
            yield uri
        else:
            raise InputNotFoundError(f"Invalid path or file: {inputs}")

    elif isinstance(inputs, Iterable):
        for item in inputs:
            if isinstance(item, (str, os.PathLike)):
                uri = _filepath_from_path_or_filelike(item)
                if is_file(uri):
                    yield uri
                elif is_directory(uri):
                    adapter = get_adapter(uri)
                    if adapter:
                        yield from process_expansion(
                            adapter.list_directory(uri, recursive=True)
                        )
                    else:
                        raise AdapterError(f"No adapter found for {uri}")
                else:
                    raise InputNotFoundError(f"Invalid path or file: {item}")
            else:
                yield item

    else:
        yield inputs



import logging
from collections import defaultdict
from datetime import datetime, timedelta
from typing import IO, Callable, Dict, NamedTuple, Optional, Set, Union

from lxml import etree, objectify

from kloppy.domain import (
    AttackingDirection,
    BallState,
    DatasetFlag,
    Metadata,
    Orientation,
    PlayerData,
    Point,
    Point3D,
    Provider,
    TrackingDataset,
    attacking_direction_from_frame,
)
from kloppy.domain.services.frame_factory import create_frame
from kloppy.infra.serializers.event.sportec.deserializer import (
    sportec_metadata_from_xml_elm,
)
from kloppy.utils import performance_logging

from ..deserializer import TrackingDataDeserializer

logger = logging.getLogger(__name__)

BALL_STATUS = "BallStatus"
BALL_POSSESSION = "BallPossession"

GAME_SECTION_TO_PERIOD_ID = {
    "firstHalf": 1,
    "secondHalf": 2,
    "firstHalfExtra": 3,
    "secondHalfExtra": 4,
}


def elem2dict(node):
    """Convert an lxml.etree node tree into a dict."""
    return dict(node.items())


def any_of(*predicates: Callable[..., bool]) -> Callable[..., bool]:
    """
    Return a new function that is True if ANY of the given predicates are True.
    """

    def combined(*args, **kwargs) -> bool:
        return any(pred(*args, **kwargs) for pred in predicates)

    return combined


def all_of(*predicates: Callable[..., bool]) -> Callable[..., bool]:
    """
    Return a new function that is True if ALL of the given predicates are True.
    """

    def combined(*args, **kwargs) -> bool:
        return all(pred(*args, **kwargs) for pred in predicates)

    return combined


def _unstack_framesets(
    raw_data: IO[bytes],
    limit: Optional[int] = None,
    only_alive: bool = True,
    objects_to_skip: Optional[Set[str]] = None,
) -> Dict[int, Dict[int, Dict]]:
    """Unstack framesets.

    Sportec groups frames per period and object in a frameset. This function
    unstacks the framesets and returns a nested dict:

        { period: { frame_id: { object_id: {attr: value, ...}, ...}, ...}, ... }

    Args:
        raw_data: The raw XML data stream to be parsed.
        limit: Max frames to collect.
        only_alive: Only include frames where BALL_STATUS == "1".
        objects_to_skip: A set of object IDs to skip.

    Returns:
        A nested dict period → frame_id → object_id → attributes.

    Notes:
        This function assumes that the framesets are ordered by period and
        that the frames in each frameset are ordered.
    """
    objects_to_skip = objects_to_skip or set()
    frames: Dict[int, Dict[int, Dict[str, Dict[str, str]]]] = defaultdict(
        lambda: defaultdict(dict)
    )
    frames_per_obj: Dict[str, int] = defaultdict(int)

    def _make_pass(
        skip_frameset: Optional[Callable[[int, str], bool]] = None,
        end_frameset: Optional[Callable[[int, str], bool]] = None,
        skip_frame: Optional[
            Callable[[int, str, Dict[str, str]], bool]
        ] = None,
    ) -> None:
        """Generic pass over the XML to collect frames.

        Args:
            skip_framset: A callable that returns True when the parsing should skip a frameset.
            end_frameset: A callable that returns True when the frameset should be ended.
            skip_frame: A callable that returns True when a frame should be skipped.
        """
        raw_data.seek(0)
        context = etree.iterparse(
            raw_data, events=("start", "end"), huge_tree=True
        )
        current_period = None
        current_obj = None

        for event, elem in context:
            if event == "start" and elem.tag == "FrameSet":
                section = elem.get("GameSection")
                current_period = GAME_SECTION_TO_PERIOD_ID.get(section)
                if current_period is None:
                    logger.warning(
                        "Found FrameSet with unknown or missing period: %s",
                        elem2dict(elem),
                    )
                    continue
                current_obj = (
                    "ball"
                    if elem.get("TeamId") == "BALL"
                    else elem.get("PersonId")
                )
                if current_obj is None:
                    logger.warning(
                        "Found FrameSet with unknown or missing object ID: %s",
                        elem2dict(elem),
                    )
                    continue
                if skip_frameset is not None and skip_frameset(
                    current_period, current_obj
                ):
                    current_period = None
                    current_obj = None
                    continue
                logger.debug(
                    "Processing FrameSet for object %s in period %s",
                    current_obj,
                    current_period,
                )

            elif (
                event == "start"
                and elem.tag == "Frame"
                and current_period is not None
                and current_obj is not None
            ):
                frame_data = elem2dict(elem)
                if skip_frame is not None and skip_frame(
                    current_period, current_obj, frame_data
                ):
                    continue
                if "N" not in frame_data:
                    logger.warning(
                        "Found Frame with missing ID: %s",
                        frame_data,
                    )
                    continue
                fid = int(frame_data["N"])
                frames[current_period][fid][current_obj] = elem2dict(elem)
                frames_per_obj[current_obj] += 1
                if end_frameset is not None and end_frameset(
                    current_period, current_obj
                ):
                    current_period = None
                    current_obj = None

            # Always clear elements to save memory
            elif event == "end" and elem.tag in ("Frame", "FrameSet"):
                elem.clear()
                while elem.getprevious() is not None:
                    del elem.getparent()[0]

    def make_object_filter(
        objects: Set[str],
        include: bool = True,
    ) -> Callable[[int, str], bool]:
        """
        Returns a callable that checks if the current object should be skipped
        based on membership in the given set.

        Args:
            objects: Set of object names to include/exclude.
            include: If True, only allow objects IN the set. If False, exclude objects IN the set.
        """

        def _filter(current_period: int, current_obj: str) -> bool:
            skip = (
                (current_obj not in objects)
                if include
                else (current_obj in objects)
            )
            if skip:
                logger.debug(
                    "Skipping FrameSet for object %s in period %s (reason: %s objects=%s)",
                    current_obj,
                    current_period,
                    ("not in" if include else "in"),
                    objects,
                )
            return skip

        return _filter

    def check_limit(current_period: int, current_obj: str) -> bool:
        """
        Check if the limit has been reached for the current object.
        """
        assert limit is not None
        skip = frames_per_obj.get(current_obj, 0) >= limit
        if skip:
            logger.debug(
                "Reached limit=%d for object %s in period %s",
                frames_per_obj.get(current_obj, 0),
                current_obj,
                current_period,
            )
        return skip

    def check_only_alive(
        current_period: int, current_obj: str, frame_data: Dict[str, str]
    ) -> bool:
        """
        Check if the ball is out of play in the current frame.
        """
        assert only_alive is True
        fid = int(frame_data.get("N", 0))

        if current_obj == "ball":
            skip = int(frame_data.get(BALL_STATUS, "0")) != 1
        else:
            skip = "ball" not in frames.get(current_period, {}).get(fid, {})

        if skip:
            logger.debug(
                "Skipping Frame %d for object %s in period %s (reason: ball not in play)",
                fid,
                current_obj,
                current_period,
            )
        return skip

    # --- Build predicates dynamically ---

    skip_frameset = make_object_filter(objects_to_skip, include=False)
    skip_frameset_ball = make_object_filter({"ball"}, include=True)
    skip_frameset_players = make_object_filter(
        objects_to_skip | {"ball"}, include=False
    )

    if limit:
        skip_frameset = any_of(skip_frameset, check_limit)
        skip_frameset_ball = any_of(skip_frameset_ball, check_limit)
        skip_frameset_players = any_of(skip_frameset_players, check_limit)

    end_frameset = check_limit if limit else None
    skip_frame = check_only_alive if only_alive else None

    # --- Processing passes ---

    if only_alive:
        # First pass: only the ball to get the frames where the ball is alive
        _make_pass(
            skip_frameset=skip_frameset_ball,
            end_frameset=end_frameset,
            skip_frame=skip_frame,
        )
        # Second pass: all other objects in these frames
        _make_pass(
            skip_frameset=skip_frameset_players,
            end_frameset=end_frameset,
            skip_frame=skip_frame,
        )
    else:
        _make_pass(
            skip_frameset=skip_frameset,
            end_frameset=end_frameset,
            skip_frame=skip_frame,
        )

    return dict(frames)


class SportecTrackingDataInputs(NamedTuple):
    meta_data: IO[bytes]
    raw_data: IO[bytes]


class SportecTrackingDataDeserializer(TrackingDataDeserializer):
    @property
    def provider(self) -> Provider:
        return Provider.SPORTEC

    def __init__(
        self,
        limit: Optional[int] = None,
        sample_rate: Optional[float] = None,
        coordinate_system: Optional[Union[str, Provider]] = None,
        only_alive: bool = True,
    ):
        super().__init__(limit, sample_rate, coordinate_system)
        self.only_alive = only_alive

    def deserialize(
        self, inputs: SportecTrackingDataInputs
    ) -> TrackingDataset:
        with performance_logging("parse metadata", logger=logger):
            match_root = objectify.fromstring(inputs.meta_data.read())
            sportec_metadata = sportec_metadata_from_xml_elm(match_root)
            date = datetime.fromisoformat(
                match_root.MatchInformation.General.attrib["KickoffTime"]
            )
            game_week = match_root.MatchInformation.General.attrib["MatchDay"]
            game_id = match_root.MatchInformation.General.attrib["MatchId"]
            periods = sportec_metadata.periods
            teams = home_team, away_team = sportec_metadata.teams
            player_map = {
                player.player_id: player
                for team in teams
                for player in team.players
            }
            official_ids = {
                official.official_id
                for official in (sportec_metadata.officials or [])
            }

            transformer = self.get_transformer(
                pitch_length=sportec_metadata.x_max,
                pitch_width=sportec_metadata.y_max,
            )

        # Stream and process tracking data
        with performance_logging("parse tracking data", logger=logger):
            period_frames = _unstack_framesets(
                inputs.raw_data,
                limit=(
                    int(self.limit / self.sample_rate)
                    if self.sample_rate and self.limit
                    else self.limit
                ),
                only_alive=self.only_alive,
                objects_to_skip=official_ids,
            )

            frames = []
            sample = 1.0 / self.sample_rate if self.sample_rate else 1
            frame_count = 0

            for period in periods:
                raw_period_frames = period_frames.get(period.id, {})

                sorted_frame_ids = sorted(raw_period_frames.keys())
                for (
                    i,
                    frame_id,
                ) in enumerate(sorted_frame_ids):
                    if self.limit and frame_count >= self.limit:
                        break

                    frame_data = raw_period_frames[frame_id]
                    if "ball" not in frame_data:
                        continue

                    ball_data = frame_data.get("ball", {})
                    if (
                        self.only_alive
                        and int(ball_data.get(BALL_STATUS, 0)) != 1
                    ):
                        continue

                    if i % sample != 0:
                        continue

                    try:
                        frame = create_frame(
                            frame_id=frame_id,
                            timestamp=timedelta(
                                seconds=(
                                    frame_id
                                    - period.start_timestamp.seconds
                                    * sportec_metadata.fps
                                )
                                / sportec_metadata.fps
                            ),
                            ball_owning_team=(
                                home_team
                                if int(ball_data.get(BALL_POSSESSION, 0)) == 1
                                else away_team
                            ),
                            ball_state=(
                                BallState.ALIVE
                                if int(ball_data.get(BALL_STATUS, 0)) == 1
                                else BallState.DEAD
                            ),
                            period=period,
                            players_data={
                                player_map[object_id]: PlayerData(
                                    coordinates=Point(
                                        x=float(data.get("X", 0)),
                                        y=float(data.get("Y", 0)),
                                    ),
                                    speed=float(data.get("S", 0)),
                                )
                                for object_id, data in frame_data.items()
                                if object_id != "ball"
                                and object_id not in official_ids
                                and object_id in player_map
                            },
                            ball_coordinates=Point3D(
                                x=float(ball_data.get("X", 0)),
                                y=float(ball_data.get("Y", 0)),
                                z=float(ball_data.get("Z", 0)),
                            ),
                            ball_speed=float(ball_data.get("S", 0)),
                            other_data={},
                        )
                        frames.append(transformer.transform_frame(frame))
                        frame_count += 1
                    except KeyError as e:
                        logger.warning(
                            f"Skipping frame {frame_id} due to missing data: {e}"
                        )

        # Determine orientation
        try:
            first_frame = next(
                frame for frame in frames if frame.period.id == 1
            )
            orientation = (
                Orientation.HOME_AWAY
                if attacking_direction_from_frame(first_frame)
                == AttackingDirection.LTR
                else Orientation.AWAY_HOME
            )
        except StopIteration:
            orientation = Orientation.NOT_SET
            logger.warning("Could not determine dataset orientation")

        metadata = Metadata(
            teams=teams,
            periods=periods,
            pitch_dimensions=transformer.get_to_coordinate_system().pitch_dimensions,
            score=sportec_metadata.score,
            frame_rate=sportec_metadata.fps,
            orientation=orientation,
            provider=Provider.SPORTEC,
            flags=DatasetFlag.BALL_OWNING_TEAM | DatasetFlag.BALL_STATE,
            coordinate_system=transformer.get_to_coordinate_system(),
            date=date,
            game_week=game_week,
            game_id=game_id,
            home_coach=sportec_metadata.home_coach,
            away_coach=sportec_metadata.away_coach,
            officials=sportec_metadata.officials,
        )

        return TrackingDataset(
            records=frames,
            metadata=metadata,
        )



import fsspec

from kloppy.config import get_config
from kloppy.exceptions import AdapterError

from .fsspec import FSSpecAdapter


class S3Adapter(FSSpecAdapter):
    def supports(self, url: str) -> bool:
        return url.startswith("s3://")

    def _get_filesystem(
        self, url: str, no_cache: bool = False
    ) -> fsspec.AbstractFileSystem:
        try:
            import s3fs
        except ImportError:
            raise AdapterError(
                "Seems like you don't have s3fs installed. Please"
                " install it using: pip install s3fs"
            )

        s3_fs = get_config("adapters.s3.s3fs") or s3fs.S3FileSystem()

        if no_cache:
            return s3_fs
        return fsspec.filesystem(
            "simplecache",
            fs=s3_fs,
            cache_storage=get_config("cache"),
        )



import dataclasses
import warnings
from dataclasses import fields

from kloppy.domain import Frame


def create_frame(**kwargs) -> Frame:
    """
    Do the actual construction of a frame.

    This method does a couple of things:
    1. Fill in some arguments when not passed
    2. Pass only arguments that are accepted by the Frame class.
    """
    if "statistics" not in kwargs:
        kwargs["statistics"] = []

    relevant_kwargs = {
        field.name: kwargs.get(field.name, field.default)
        for field in fields(Frame)
        if field.init
        and not (
            field.default == dataclasses.MISSING and field.name not in kwargs
        )
    }

    if len(relevant_kwargs) < len(kwargs):
        skipped_kwargs = set(kwargs.keys()) - set(relevant_kwargs.keys())
        warnings.warn(
            f"The following arguments were skipped: {skipped_kwargs}"
        )

    frame = Frame(**relevant_kwargs)

    return frame



from kloppy.domain import CodeDataset
from kloppy.infra.serializers.code.sportscode import (
    SportsCodeDeserializer,
    SportsCodeInputs,
    SportsCodeSerializer,
)
from kloppy.io import FileLike, open_as_file


def load(data: FileLike) -> CodeDataset:
    """
    Load SportsCode data.

    Args:
        data: The raw SportsCode XML data.

    Returns:
        The parsed SportsCode data.
    """
    deserializer = SportsCodeDeserializer()

    with open_as_file(data) as data_fp:
        return deserializer.deserialize(inputs=SportsCodeInputs(data=data_fp))


def save(dataset: CodeDataset, output_filename: str) -> None:
    """
    Save SportsCode data to an XML file.

    Args:
        dataset: The SportsCode dataset to save.
        output_filename: The output filename.
    """
    with open(output_filename, "wb") as fp:
        serializer = SportsCodeSerializer()
        fp.write(serializer.serialize(dataset))



import re
from abc import ABC, abstractmethod
from typing import BinaryIO, List, Optional

import fsspec

from kloppy.config import get_config
from kloppy.exceptions import InputNotFoundError

from .adapter import Adapter


class FSSpecAdapter(Adapter, ABC):
    def _infer_protocol(self, url: str) -> str:
        """
        Infer the protocol based on the URL prefix.
        """
        protocol_pattern = re.compile(r"^[a-zA-Z\d]+://")
        match = protocol_pattern.match(url)
        if match:
            return match.group(0)[
                :-3
            ]  # Remove '://' from the matched protocol
        return "file"  # Default to 'file' for local paths

    def _get_filesystem(
        self, url: str, no_cache: bool = False
    ) -> fsspec.AbstractFileSystem:
        """
        Get the appropriate fsspec filesystem for the given URL, with caching enabled.
        """
        protocol = self._infer_protocol(url)

        if no_cache:
            return fsspec.filesystem(protocol)

        return fsspec.filesystem(
            "simplecache",
            target_protocol=protocol,
            cache_storage=get_config("cache"),
        )

    def _detect_compression(self, url: str) -> Optional[str]:
        """
        Detect the compression type based on the file extension.
        Supported compression types include 'gzip', 'bz2', 'xz', etc.
        """
        compression_map = {
            ".gz": "gzip",
            ".bz2": "bz2",
            ".xz": "xz",
            ".zip": "zip",
        }
        for ext, comp in compression_map.items():
            if url.endswith(ext):
                return comp
        return None  # No compression detected

    @abstractmethod
    def supports(self, url: str) -> bool:
        """
        Check if the adapter can handle the URL.
        """

    def read_to_stream(self, url: str, output: BinaryIO):
        """
        Reads content from the given URL and writes it to the provided binary stream.
        Uses caching for remote files.
        """
        fs = self._get_filesystem(url)
        compression = self._detect_compression(url)

        try:
            with fs.open(url, "rb", compression=compression) as source_file:
                output.write(source_file.read())
        except FileNotFoundError as e:
            raise InputNotFoundError(f"Input file not found: {url}") from e

    def list_directory(self, url: str, recursive: bool = True) -> List[str]:
        """
        Lists the contents of a directory.
        """
        protocol = self._infer_protocol(url)
        fs = self._get_filesystem(url)
        if recursive:
            files = fs.find(url, detail=False)
        else:
            files = fs.listdir(url, detail=False)
        return [
            f"{protocol}://{fp}"
            if protocol != "file" and not fp.startswith(protocol)
            else fp
            for fp in files
        ]

    def is_directory(self, url: str) -> bool:
        """
        Check if the given URL points to a directory.
        """
        fs = self._get_filesystem(url)
        return fs.isdir(url)

    def is_file(self, url: str) -> bool:
        """
        Check if the given URL points to a file.
        """
        fs = self._get_filesystem(url)
        return fs.isfile(url)



import logging
import warnings
from typing import IO, NamedTuple, Optional, Union

from kloppy.domain import (
    AttackingDirection,
    DatasetFlag,
    Metadata,
    Orientation,
    Provider,
    TrackingDataset,
    attacking_direction_from_frame,
)
from kloppy.utils import performance_logging

from ..deserializer import TrackingDataDeserializer
from .parsers import get_metadata_parser, get_raw_data_parser

logger = logging.getLogger(__name__)


class TRACABInputs(NamedTuple):
    meta_data: IO[bytes]
    raw_data: IO[bytes]


class TRACABDeserializer(TrackingDataDeserializer[TRACABInputs]):
    def __init__(
        self,
        limit: Optional[int] = None,
        sample_rate: Optional[float] = None,
        coordinate_system: Optional[Union[str, Provider]] = None,
        only_alive: bool = True,
    ):
        super().__init__(limit, sample_rate, coordinate_system)
        self.only_alive = only_alive

    @property
    def provider(self) -> Provider:
        return Provider.TRACAB

    def deserialize(self, inputs: TRACABInputs) -> TrackingDataset:
        with performance_logging("Loading metadata", logger=logger):
            metadata_parser = get_metadata_parser(inputs.meta_data)
            (
                pitch_length,
                pitch_width,
            ) = metadata_parser.extract_pitch_dimensions()
            teams = metadata_parser.extract_lineups()
            periods = metadata_parser.extract_periods()
            frame_rate = metadata_parser.extract_frame_rate()
            date = metadata_parser.extract_date()
            game_id = metadata_parser.extract_game_id()
            orientation = metadata_parser.extract_orientation()

        transformer = self.get_transformer(
            pitch_length=pitch_length, pitch_width=pitch_width
        )

        with performance_logging("Loading data", logger=logger):
            raw_data_parser = get_raw_data_parser(
                inputs.raw_data, periods, teams, frame_rate
            )
            frames = []
            for n, frame in enumerate(
                raw_data_parser.extract_frames(
                    self.sample_rate, self.only_alive
                )
            ):
                frame = transformer.transform_frame(frame)
                frames.append(frame)

                if self.limit and n + 1 >= (self.limit / self.sample_rate):
                    break

        if orientation is None:
            try:
                first_frame = next(
                    frame for frame in frames if frame.period.id == 1
                )
                orientation = (
                    Orientation.HOME_AWAY
                    if attacking_direction_from_frame(first_frame)
                    == AttackingDirection.LTR
                    else Orientation.AWAY_HOME
                )
            except StopIteration:
                warnings.warn(
                    "Could not determine orientation of dataset, defaulting to NOT_SET"
                )
                orientation = Orientation.NOT_SET

        metadata = Metadata(
            teams=list(teams),
            periods=periods,
            pitch_dimensions=transformer.get_to_coordinate_system().pitch_dimensions,
            score=None,
            frame_rate=frame_rate,
            orientation=orientation,
            provider=Provider.TRACAB,
            flags=DatasetFlag.BALL_OWNING_TEAM | DatasetFlag.BALL_STATE,
            coordinate_system=transformer.get_to_coordinate_system(),
            date=date,
            game_id=game_id,
        )

        return TrackingDataset(
            records=frames,
            metadata=metadata,
        )



from typing import IO, List, Optional, Tuple

from lxml import objectify

from kloppy.domain import Period, Team

from .metadata.base import TracabMetadataParser
from .metadata.flat_xml import TracabFlatXMLMetadataParser
from .metadata.hierarchical_xml import TracabHierarchicalXMLMetadataParser
from .metadata.json import TracabJSONMetadataParser
from .raw_data.base import TracabDataParser
from .raw_data.dat import TracabDatParser
from .raw_data.json import TracabJSONParser


def get_metadata_parser(
    feed: IO[bytes], feed_format: Optional[str] = None
) -> TracabMetadataParser:
    # infer the data format if not provided
    if feed_format is None:
        if feed.read(1).decode("utf-8")[0] == "<":
            feed.seek(0)
            meta_data = objectify.fromstring(feed.read())
            if hasattr(meta_data, "match"):
                feed_format = "HIERARCHICAL_XML"
            else:
                feed_format = "FLAT_XML"
        else:
            feed_format = "JSON"
        feed.seek(0)

    if feed_format.upper() == "JSON":
        return TracabJSONMetadataParser(feed)
    elif feed_format.upper() == "FLAT_XML":
        return TracabFlatXMLMetadataParser(feed)
    elif feed_format.upper() == "HIERARCHICAL_XML":
        return TracabHierarchicalXMLMetadataParser(feed)
    else:
        raise ValueError(f"Unknown metadata feed format {feed_format}")


def get_raw_data_parser(
    feed: IO[bytes],
    periods: List[Period],
    teams: Tuple[Team, Team],
    frame_rate: int,
    feed_format: Optional[str] = None,
) -> TracabDataParser:
    # infer the data format if not provided
    if feed_format is None:
        if feed.read(1).decode("utf-8")[0] == "{":
            feed_format = "JSON"
        else:
            feed_format = "DAT"
        feed.seek(0)

    if feed_format.upper() == "DAT":
        return TracabDatParser(feed, periods, teams, frame_rate)
    elif feed_format.upper() == "JSON":
        return TracabJSONParser(feed, periods, teams, frame_rate)
    else:
        raise ValueError(f"Unknown raw data feed format {feed_format}")



import warnings
from dataclasses import dataclass
from enum import Enum
from math import sqrt
from typing import Optional

from kloppy.exceptions import MissingDimensionError

DEFAULT_PITCH_LENGTH = 105.0
DEFAULT_PITCH_WIDTH = 68.0


class Unit(Enum):
    """
    Unit to measure distances on a pitch.

    Attributes:
        METERS (str): Meters
        YARDS (str): Yards
        CENTIMETERS (str): Centimeters
        FEET (str): Feet
        NORMED (str): Normalized to a unit square
    """

    METERS = "m"
    YARDS = "y"
    CENTIMETERS = "cm"
    FEET = "ft"
    NORMED = "normed"

    def convert(self, to_unit: "Unit", value: float) -> float:
        """Converts a value from one unit to another.

        Arguments:
            to_unit: The unit to convert to
            value: The value to convert
        Returns:
            The value converted to the target unit
        """
        conversion_factors = {
            (Unit.METERS, Unit.METERS): 1,
            (Unit.METERS, Unit.YARDS): 1.09361,
            (Unit.METERS, Unit.CENTIMETERS): 100,
            (Unit.METERS, Unit.FEET): 3.281,
        }
        if self == to_unit:
            return value
        elif self == Unit.NORMED or to_unit == Unit.NORMED:
            raise ValueError("Cannot convert to or from a normed unit")

        factor_to_meter = conversion_factors.get((Unit.METERS, self))
        factor_from_meter = conversion_factors.get((Unit.METERS, to_unit))
        assert (
            factor_to_meter is not None
        ), f"Conversion factor for {self} is not defined"
        assert (
            factor_from_meter is not None
        ), f"Conversion factor for {to_unit} is not defined"
        return value / factor_to_meter * factor_from_meter


@dataclass(frozen=True)
class Point:
    """
    Point on the pitch.

    Attributes:
        x: x coordinate in unit of [`PitchDimensions`][kloppy.domain.models.pitch.PitchDimensions]
        y: y coordinate in unit of [`PitchDimensions`][kloppy.domain.models.pitch.PitchDimensions]
    """

    x: float
    y: float

    def distance_to(self, other: "Point") -> float:
        """
        Calculates the euclidean distance between the point and another provided point

        Arguments:
            other: See [`Point`][kloppy.domain.models.pitch.Point]
        """
        return sqrt((self.x - other.x) ** 2 + (self.y - other.y) ** 2)


@dataclass(frozen=True)
class Point3D(Point):
    """
    Point on the pitch that includes the z coordinate for height (e.g. of the ball).

    Attributes:
        z: z coordinate in unit of [`PitchDimensions`][kloppy.domain.models.pitch.PitchDimensions]
    """

    z: Optional[float]


@dataclass(frozen=True)
class Dimension:
    """
    Limits of pitch boundaries along a single axis.

    Attributes:
        min: Minimal possible value within this dimension
        max: Maximal possible value within this dimension
    """

    min: Optional[float] = None
    max: Optional[float] = None

    def to_base(self, value: float) -> float:
        """Map a value from this dimension to [0, 1]."""
        if self.min is None or self.max is None:
            raise MissingDimensionError()
        return (value - self.min) / (self.max - self.min)

    def from_base(self, value: float) -> float:
        """Map a value from [0, 1] to this dimension."""
        if self.min is None or self.max is None:
            raise MissingDimensionError()
        return value * (self.max - self.min) + self.min


@dataclass
class PitchDimensions:
    """
    Specifies the dimensions of a pitch.

    Attributes:
        x_dim: Limits of pitch boundaries in longitudinal direction.
        y_dim: Limits of pitch boundaries in lateral direction.
        standardized: Used to denote standardized pitch dimensions where data
            is scaled along the axes independent of the actual dimensions of
            the pitch. To get non-distored calculations, the `length` and
            `width` of the pitch need to be specified.
        unit: The unit in which distances are measured along the axes of the pitch.
        goal_width: Width of the goal.
        goal_height: Height of the goal.
        six_yard_width: Width of the six yard box.
        six_yard_length: Length of the six yard box.
        penalty_area_width: Width of the penalty area.
        penalty_area_length: Length of the penalty area.
        circle_radius: Radius of the center circle (in the longitudinal direction).
        corner_radius: Radius of the corner arcs.
        penalty_spot_distance: Distance from the goal line to the penalty spot.
        penalty_arc_radius: Radius of the penalty arc (in the longitudinal direction).
        pitch_length: True length of the pitch, in meters.
        pitch_width: True width of the pitch, in meters.
    """

    x_dim: Dimension
    y_dim: Dimension
    standardized: bool
    unit: Unit

    goal_width: float
    goal_height: Optional[float]
    six_yard_width: float
    six_yard_length: float
    penalty_area_width: float
    penalty_area_length: float
    circle_radius: float
    corner_radius: float
    penalty_spot_distance: float
    penalty_arc_radius: float

    pitch_length: Optional[float] = None
    pitch_width: Optional[float] = None

    def convert(self, to_unit: Unit) -> "PitchDimensions":
        """Convert the pitch dimensions to another unit.

        Arguments:
            to_unit: The unit to convert to

        Returns:
            The pitch dimensions in the target unit
        """
        return PitchDimensions(
            x_dim=Dimension(
                min=(
                    self.unit.convert(to_unit, self.x_dim.min)
                    if self.x_dim.min is not None
                    else None
                ),
                max=(
                    self.unit.convert(to_unit, self.x_dim.max)
                    if self.x_dim.max is not None
                    else None
                ),
            ),
            y_dim=Dimension(
                min=(
                    self.unit.convert(to_unit, self.y_dim.min)
                    if self.y_dim.min is not None
                    else None
                ),
                max=(
                    self.unit.convert(to_unit, self.y_dim.max)
                    if self.y_dim.max is not None
                    else None
                ),
            ),
            standardized=self.standardized,
            unit=to_unit,
            goal_width=self.unit.convert(to_unit, self.goal_width),
            goal_height=(
                self.unit.convert(to_unit, self.goal_height)
                if self.goal_height is not None
                else None
            ),
            six_yard_width=self.unit.convert(to_unit, self.six_yard_width),
            six_yard_length=self.unit.convert(to_unit, self.six_yard_length),
            penalty_area_width=self.unit.convert(
                to_unit, self.penalty_area_width
            ),
            penalty_area_length=self.unit.convert(
                to_unit, self.penalty_area_length
            ),
            circle_radius=self.unit.convert(to_unit, self.circle_radius),
            corner_radius=self.unit.convert(to_unit, self.corner_radius),
            penalty_spot_distance=self.unit.convert(
                to_unit, self.penalty_spot_distance
            ),
            penalty_arc_radius=self.unit.convert(
                to_unit, self.penalty_arc_radius
            ),
            pitch_length=self.pitch_length,
            pitch_width=self.pitch_width,
        )

    def _transformation_zones_x(self, length: float):
        assert self.x_dim.min is not None
        penalty_arc = self.penalty_spot_distance + self.penalty_arc_radius
        return [
            # goal line to 6 yard box
            (self.x_dim.min, self.x_dim.min + self.six_yard_length),
            # 6 yard box to penalty spot
            (
                self.x_dim.min + self.six_yard_length,
                self.x_dim.min + self.penalty_spot_distance,
            ),
            # penalty spot to penalty area
            (
                self.x_dim.min + self.penalty_spot_distance,
                self.x_dim.min + self.penalty_area_length,
            ),
            # penalty area to penalty arc
            (
                self.x_dim.min + self.penalty_area_length,
                self.x_dim.min + penalty_arc,
            ),
            # penalty arc to center circle
            (
                self.x_dim.min + penalty_arc,
                self.x_dim.min + length / 2 - self.circle_radius,
            ),
            # center circle to center line
            (
                self.x_dim.min + length / 2 - self.circle_radius,
                self.x_dim.min + length / 2,
            ),
        ]

    def _transformation_zones_y(self, width: float):
        assert self.y_dim.min is not None
        return [
            # side line to penalty area
            (
                self.y_dim.min,
                self.y_dim.min + (width - self.penalty_area_width) / 2,
            ),
            # penalty area to six yard box
            (
                self.y_dim.min + (width - self.penalty_area_width) / 2,
                self.y_dim.min + (width - self.six_yard_width) / 2,
            ),
            # six yard box to inside goalpost
            (
                self.y_dim.min + (width - self.six_yard_width) / 2,
                self.y_dim.min + (width - self.goal_width) / 2,
            ),
            # inside goalpost to center
            (
                self.y_dim.min + (width - self.goal_width) / 2,
                self.y_dim.min + width / 2,
            ),
        ]

    def to_metric_base(
        self,
        point: Point,
        pitch_length: float = DEFAULT_PITCH_LENGTH,
        pitch_width: float = DEFAULT_PITCH_WIDTH,
    ) -> Point:
        """
        Convert a point from this pitch dimensions to the IFAB pitch dimensions.

        Arguments:
            point: The point to convert

        Returns:
            The point in the IFAB pitch dimensions
        """
        if (
            self.x_dim.min is None
            or self.x_dim.max is None
            or self.y_dim.min is None
            or self.y_dim.max is None
        ):
            raise MissingDimensionError(
                "The pitch boundaries need to be fully specified to convert coordinates."
            )

        ifab_dims = MetricPitchDimensions(
            x_dim=Dimension(0, pitch_length),
            y_dim=Dimension(0, pitch_width),
            pitch_length=pitch_length,
            pitch_width=pitch_width,
            standardized=False,
        )
        x_ifab_zones = ifab_dims._transformation_zones_x(pitch_length)
        y_ifab_zones = ifab_dims._transformation_zones_y(pitch_width)
        x_from_zones = self._transformation_zones_x(
            self.x_dim.max - self.x_dim.min
        )
        y_from_zones = self._transformation_zones_y(
            self.y_dim.max - self.y_dim.min
        )

        def transform(v, from_zones, from_length, ifab_zones, ifab_length):
            mirror = False
            if v > from_zones[-1][1]:
                v = from_length - (v - from_zones[0][0]) + from_zones[0][0]
                mirror = True
            # find the zone the v coordinate is in
            try:
                zone = next(
                    (
                        idx
                        for idx, zone in enumerate(from_zones)
                        if zone[0] <= v <= zone[1]
                    )
                )
                scale = (
                    # length of the zone in IFAB dimensions
                    (ifab_zones[zone][1] - ifab_zones[zone][0])
                    # length of the zone in the original dimensions
                    / (from_zones[zone][1] - from_zones[zone][0])
                )
                # ifab = smallest IFAB value of the zone + (v - smallest v value of the zone) * scaling factor of the zone
                ifab = ifab_zones[zone][0] + (v - from_zones[zone][0]) * scale
            except StopIteration:
                # value is outside of the pitch dimensions
                ifab = ifab_zones[0][0] + (v - from_zones[0][0]) * (
                    ifab_length / from_length
                )
            if mirror:
                ifab = ifab_length - ifab
            return ifab

        if isinstance(point, Point3D):
            return Point3D(
                x=transform(
                    point.x,
                    x_from_zones,
                    self.x_dim.max - self.x_dim.min,
                    x_ifab_zones,
                    pitch_length,
                ),
                y=transform(
                    point.y,
                    y_from_zones,
                    self.y_dim.max - self.y_dim.min,
                    y_ifab_zones,
                    pitch_width,
                ),
                z=(
                    (
                        point.z * 2.44 / self.goal_height
                        if self.goal_height is not None
                        else point.z
                    )
                    if point.z is not None
                    else None
                ),
            )
        else:
            return Point(
                x=transform(
                    point.x,
                    x_from_zones,
                    self.x_dim.max - self.x_dim.min,
                    x_ifab_zones,
                    pitch_length,
                ),
                y=transform(
                    point.y,
                    y_from_zones,
                    self.y_dim.max - self.y_dim.min,
                    y_ifab_zones,
                    pitch_width,
                ),
            )

    def from_metric_base(
        self,
        point: Point,
        pitch_length: float = DEFAULT_PITCH_LENGTH,
        pitch_width: float = DEFAULT_PITCH_WIDTH,
    ) -> Point:
        """
        Convert a point from the IFAB pitch dimensions to this pitch dimensions.

        Arguments:
            point: The point to convert

        Returns:
            The point in the regular pitch dimensions
        """
        if (
            self.x_dim.min is None
            or self.x_dim.max is None
            or self.y_dim.min is None
            or self.y_dim.max is None
        ):
            raise MissingDimensionError(
                "The pitch boundaries need to be fully specified to convert coordinates."
            )

        ifab_dims = MetricPitchDimensions(
            x_dim=Dimension(0, pitch_length),
            y_dim=Dimension(0, pitch_width),
            pitch_length=pitch_length,
            pitch_width=pitch_width,
            standardized=False,
        )
        x_ifab_zones = ifab_dims._transformation_zones_x(pitch_length)
        y_ifab_zones = ifab_dims._transformation_zones_y(pitch_width)
        x_to_zones = self._transformation_zones_x(
            self.x_dim.max - self.x_dim.min
        )
        y_to_zones = self._transformation_zones_y(
            self.y_dim.max - self.y_dim.min
        )

        def transform(v, to_zones, to_length, ifab_zones, ifab_length):
            mirror = False
            if v > ifab_length / 2:
                v = ifab_length - v
                mirror = True
            # find the zone the v coordinate is in
            try:
                zone = next(
                    (
                        idx
                        for idx, zone in enumerate(ifab_zones)
                        if zone[0] <= v <= zone[1]
                    )
                )
                scale = (
                    # length of the zone in the original dimensions
                    (to_zones[zone][1] - to_zones[zone][0])
                    # length of the zone in IFAB dimensions
                    / (ifab_zones[zone][1] - ifab_zones[zone][0])
                )
                # ifab = smallest IFAB value of the zone + (v - smallest v value of the zone) * scaling factor of the zone
                v = to_zones[zone][0] + (v - ifab_zones[zone][0]) * scale
            except StopIteration:
                # value is outside of the pitch dimensions
                v = to_zones[0][0] + (v - ifab_zones[0][0]) * (
                    to_length / ifab_length
                )
            if mirror:
                v = (to_length + to_zones[0][0] - v) + to_zones[0][0]
            return v

        if isinstance(point, Point3D):
            return Point3D(
                x=transform(
                    point.x,
                    x_to_zones,
                    self.x_dim.max - self.x_dim.min,
                    x_ifab_zones,
                    pitch_length,
                ),
                y=transform(
                    point.y,
                    y_to_zones,
                    self.y_dim.max - self.y_dim.min,
                    y_ifab_zones,
                    pitch_width,
                ),
                z=(
                    (
                        point.z * self.goal_height / 2.44
                        if self.goal_height is not None
                        else point.z
                    )
                    if point.z is not None
                    else None
                ),
            )
        else:
            return Point(
                x=transform(
                    point.x,
                    x_to_zones,
                    self.x_dim.max - self.x_dim.min,
                    x_ifab_zones,
                    pitch_length,
                ),
                y=transform(
                    point.y,
                    y_to_zones,
                    self.y_dim.max - self.y_dim.min,
                    y_ifab_zones,
                    pitch_width,
                ),
            )

    def distance_between(
        self, point1: Point, point2: Point, unit: Unit = Unit.METERS
    ) -> float:
        """
        Calculate the distance between two points in the coordinate system.

        Arguments:
            point1: The first point
            point2: The second point
            unit: The unit to measure the distance in

        Returns:
            The distance between the two points in the given unit
        """
        if self.pitch_length is None or self.pitch_width is None:
            warnings.warn(
                "The pitch length and width are not specified. "
                "Assuming a standard pitch size of 105x68 meters. "
                "This may lead to incorrect results.",
                stacklevel=2,
            )
            pitch_length = DEFAULT_PITCH_LENGTH
            pitch_width = DEFAULT_PITCH_WIDTH
        else:
            pitch_length = self.pitch_length
            pitch_width = self.pitch_width
        point1_ifab = self.to_metric_base(point1, pitch_length, pitch_width)
        point2_ifab = self.to_metric_base(point2, pitch_length, pitch_width)
        dist = point1_ifab.distance_to(point2_ifab)
        return Unit.METERS.convert(unit, dist)


@dataclass
class MetricPitchDimensions(PitchDimensions):
    """
    The standard pitch dimensions in meters by IFAB regulations.

    For national matches, the length of the pitch can be between 90 and 120
    meters, and the width can be between 45 and 90 meters. For international
    matches, the length can be between 100 and 110 meters, and the width can
    be between 64 and 75 meters. All other dimensions are fixed.

    See https://www.theifab.com/laws/latest/the-field-of-play.
    """

    unit: Unit = Unit.METERS

    goal_width: float = 7.32
    goal_height: Optional[float] = 2.44
    six_yard_width: float = 18.32
    six_yard_length: float = 5.5
    penalty_area_width: float = 40.32
    penalty_area_length: float = 16.5
    circle_radius: float = 9.15
    corner_radius: float = 1
    penalty_spot_distance: float = 11
    penalty_arc_radius: float = 9.15


@dataclass
class ImperialPitchDimensions(PitchDimensions):
    """
    The standard pitch dimensions in yards by IFAB regulations.

    For national matches, the length of the pitch can be between 100 and 130
    yards, and the width can be between 50 and 100 yards. For international
    matches, the length can be between 110 and 120 yards, and the width can
    be between 70 and 80 yards. All other dimensions are fixed.

    See https://www.theifab.com/laws/latest/the-field-of-play.
    """

    unit: Unit = Unit.YARDS

    goal_width: float = 8
    goal_height: Optional[float] = 2.66
    six_yard_width: float = 20
    six_yard_length: float = 6
    penalty_area_width: float = 44
    penalty_area_length: float = 18
    circle_radius: float = 10
    corner_radius: float = 1
    penalty_spot_distance: float = 12
    penalty_arc_radius: float = 10


@dataclass
class NormalizedPitchDimensions(MetricPitchDimensions):
    """
    The standard pitch dimensions in normalized units.

    The pitch dimensions are normalized to a unit square, where the length
    and width of the pitch are 1. All other dimensions are scaled accordingly.
    """

    x_dim: Dimension = Dimension(0, 1)
    y_dim: Dimension = Dimension(0, 1)
    standardized: bool = False
    unit: Unit = Unit.NORMED

    def __post_init__(self):
        if self.pitch_length is None or self.pitch_width is None:
            raise ValueError("The pitch length and width need to be specified")
        if (
            self.x_dim.min is None
            or self.x_dim.max is None
            or self.y_dim.min is None
            or self.y_dim.max is None
        ):
            raise ValueError(
                "The pitch dimensions need to be fully specified."
            )
        dim_width = self.y_dim.max - self.y_dim.min
        dim_length = self.x_dim.max - self.x_dim.min
        self.goal_width = self.goal_width / self.pitch_width * dim_width
        self.six_yard_width = (
            self.six_yard_width / self.pitch_width * dim_width
        )
        self.six_yard_length = (
            self.six_yard_length / self.pitch_length * dim_length
        )
        self.penalty_area_width = (
            self.penalty_area_width / self.pitch_width * dim_width
        )
        self.penalty_area_length = (
            self.penalty_area_length / self.pitch_length * dim_length
        )
        self.circle_radius = (
            self.circle_radius / self.pitch_length * dim_length
        )
        self.corner_radius = (
            self.corner_radius / self.pitch_length * dim_length
        )
        self.penalty_spot_distance = (
            self.penalty_spot_distance / self.pitch_length * dim_length
        )
        self.penalty_arc_radius = (
            self.penalty_arc_radius / self.pitch_length * dim_length
        )


@dataclass
class OptaPitchDimensions(PitchDimensions):
    """The pitch dimensions used by Opta."""

    x_dim: Dimension = Dimension(0, 100)
    y_dim: Dimension = Dimension(0, 100)
    standardized: bool = True
    unit: Unit = Unit.NORMED

    goal_width: float = 9.6
    goal_height: Optional[float] = 38
    six_yard_width: float = 26.4
    six_yard_length: float = 5.8
    penalty_area_width: float = 57.8
    penalty_area_length: float = 17.0
    circle_radius: float = 9.00
    corner_radius: float = 0.97
    penalty_spot_distance: float = 11.5
    penalty_arc_radius: float = 8.9


@dataclass
class WyscoutPitchDimensions(PitchDimensions):
    """The pitch dimensions used by Wyscout."""

    x_dim: Dimension = Dimension(0, 100)
    y_dim: Dimension = Dimension(0, 100)
    standardized: bool = True
    unit: Unit = Unit.NORMED

    goal_width: float = 12.0
    goal_height: Optional[float] = None
    six_yard_width: float = 26.0
    six_yard_length: float = 6.0
    penalty_area_width: float = 62.0
    penalty_area_length: float = 16.0
    circle_radius: float = 8.84  # inferred
    corner_radius: float = 0.97  # inferred
    penalty_spot_distance: float = 10.0
    penalty_arc_radius: float = 7.74  # inferred



import functools
import inspect
import re
import time
import warnings
from contextlib import contextmanager
from io import BytesIO
from logging import Logger
from typing import BinaryIO, Optional, Union

Readable = Union[bytes, BinaryIO]


def to_file_object(s: Readable) -> BinaryIO:
    if isinstance(s, bytes):
        return BytesIO(s)
    return s


@contextmanager
def performance_logging(
    description: str,
    counter: Optional[int] = None,
    logger: Optional[Logger] = None,
):
    start = time.time()
    try:
        yield
    finally:
        took = (time.time() - start) * 1000
        extra = ""
        if counter is not None:
            extra = f" ({int(counter / took * 1000)}items/sec)"

        unit = "ms"
        if took < 0.1:
            took *= 1000
            unit = "us"

        msg = f"{description} took: {took:.2f}{unit} {extra}"
        if logger:
            logger.info(msg)
        else:
            print(msg)


_first_cap_re = re.compile("(.)([A-Z][a-z0-9]+)")
_all_cap_re = re.compile("([a-z0-9])([A-Z])")


def camelcase_to_snakecase(name: str) -> str:
    """Convert camel-case string to snake-case."""
    s1 = _first_cap_re.sub(r"\1_\2", name)
    return _all_cap_re.sub(r"\1_\2", s1).lower()


def removes_suffix(string: str, suffix: str) -> str:
    if string[-len(suffix) :] == suffix:
        return string[: -len(suffix)]
    else:
        return string


def docstring_inherit_attributes(parent):
    def inherit(obj):
        other_docs, attribute_docs = obj.__doc__.split("Attributes:\n")

        own_attributes = [
            attribute.strip()
            for attribute in attribute_docs.strip().split("\n")
        ]

        parent_attributes = [
            attribute.strip()
            for attribute in parent.__doc__.split("Attributes:\n")[-1]
            .strip()
            .split("\n")
        ]
        obj.__doc__ = (
            other_docs
            + "Attributes:\n        "
            + "\n        ".join(parent_attributes)
            + "\n        "
            + "\n        ".join(own_attributes)
            + "\n"
        )
        return obj

    return inherit


string_types = (bytes, str)


def deprecated(reason):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    if isinstance(reason, string_types):
        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #      pass

        def decorator(func1):
            if inspect.isclass(func1):
                fmt1 = "Call to deprecated class {name} ({reason})."
            else:
                fmt1 = "Call to deprecated function {name} ({reason})."

            @functools.wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.simplefilter("always", DeprecationWarning)
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason),
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                warnings.simplefilter("default", DeprecationWarning)
                return func1(*args, **kwargs)

            return new_func1

        return decorator

    elif inspect.isclass(reason) or inspect.isfunction(reason):
        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #      pass

        func2 = reason

        if inspect.isclass(func2):
            fmt2 = "Call to deprecated class {name}."
        else:
            fmt2 = "Call to deprecated function {name}."

        @functools.wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(
                fmt2.format(name=func2.__name__),
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)
            return func2(*args, **kwargs)

        return new_func2

    else:
        raise TypeError(repr(type(reason)))


class DeprecatedEnumValue:
    def __init__(self, value):
        self.value = value

    def __get__(self, instance, owner):
        warnings.warn(
            f"{owner.__name__} is deprecated. Use GoalkeeperActionType instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.value


def snake_case(s: str) -> str:
    """Convert a string to snake_case."""
    return re.sub(r"[\s\-]+", "_", s.strip()).lower()



"""Base class for all parsers that can handle Tracab raw data files.

A parser reads a single data file and should extend the 'TracabDataParser'
class to extract the tracking data frames.
"""

from abc import ABC, abstractmethod
from typing import IO, Iterator, List, Tuple

from kloppy.domain import Frame, Period, Team


class TracabDataParser(ABC):
    """Extract data from a Tracab tracking data stream."""

    def __init__(
        self,
        feed: IO[bytes],
        periods: List[Period],
        teams: Tuple[Team, Team],
        frame_rate: int,
    ) -> None:
        """Initialize the parser with the data stream and metadata.

        Args:
            feed : The data stream of a game to parse.
            periods : The periods of the game.
            teams : The home and away teams.
            frame_rate : The frame rate of the tracking data.
        """
        self.periods = periods
        self.teams = teams
        self.frame_rate = frame_rate

    @abstractmethod
    def extract_frames(
        self, sample_rate: float, only_alive: bool
    ) -> Iterator[Frame]:
        """Extract all frames."""



from .formation import FormationStateBuilder
from .lineup import LineupStateBuilder
from .score import ScoreStateBuilder
from .sequence import SequenceStateBuilder

__all__ = [
    "LineupStateBuilder",
    "ScoreStateBuilder",
    "SequenceStateBuilder",
    "FormationStateBuilder",
]



"""Functions for loading Metrica Sports event and tracking data."""

from ._providers.metrica import (
    load_event,
    load_open_data,
    load_tracking_csv,
    load_tracking_epts,
)

__all__ = [
    "load_event",
    "load_tracking_csv",
    "load_tracking_epts",
    "load_open_data",
]
